from datetime import datetime


def get_wiki_tool():
    from langchain_community.tools.wikidata.tool import (
        WikidataAPIWrapper,
        WikidataQueryRun,
    )

    wiki = WikidataQueryRun(api_wrapper=WikidataAPIWrapper())  # type: ignore
    wiki.description = (
        "Look up information on a specific string from Wikipedia. Use to get context"
    )
    return wiki


def get_today_date(query: str) -> str:
    """
    Useful to get the date of today.
    """
    # Getting today's date in string format
    today_date_string = datetime.now().strftime("%Y-%m-%d")
    return today_date_string



from logging import DEBUG, StreamHandler
from os.path import dirname
from pathlib import PurePath
from typing import Optional

import pandas as pd
from sqlalchemy import create_engine
from trilogy import Dialects, Executor
from trilogy.core.enums import FunctionType, Purpose
from trilogy.core.functions import arg_to_datatype, function_args_to_output_purpose
from trilogy.core.models import (
    ColumnAssignment,
    Concept,
    Datasource,
    DataType,
    Environment,
    Function,
    Grain,
)

from trilogy_nlp.constants import logger
from trilogy_nlp.main import build_query
from trilogy_nlp.main_v2 import build_query as build_query_v2


def create_function_derived_concept(
    name: str,
    namespace: str,
    operator: FunctionType,
    arguments: list[Concept],
    output_type: Optional[DataType] = None,
    output_purpose: Optional[Purpose] = None,
) -> Concept:
    purpose = (
        function_args_to_output_purpose(arguments)
        if output_purpose is None
        else output_purpose
    )
    output_type = arg_to_datatype(arguments[0]) if output_type is None else output_type
    return Concept(
        name=name,
        namespace=namespace,
        datatype=output_type,
        purpose=purpose,
        lineage=Function(
            operator=operator,
            arguments=arguments,
            output_datatype=output_type,
            output_purpose=purpose,
            arg_count=len(arguments),
        ),
    )


def setup_engine() -> Executor:
    engine = create_engine(r"duckdb:///:memory:", future=True)
    csv = PurePath(dirname(__file__)) / "train.csv"
    df = pd.read_csv(csv)
    _ = df
    output = Executor(engine=engine, dialect=Dialects.DUCK_DB)

    output.execute_raw_sql("CREATE TABLE raw_titanic AS SELECT * FROM df")
    return output


def setup_titanic(env: Environment):
    namespace = "passenger"
    id = Concept(
        name="id", namespace=namespace, datatype=DataType.INTEGER, purpose=Purpose.KEY
    )
    age = Concept(
        name="age",
        namespace=namespace,
        datatype=DataType.INTEGER,
        purpose=Purpose.PROPERTY,
        keys=[id],
    )

    name = Concept(
        name="name",
        namespace=namespace,
        datatype=DataType.STRING,
        purpose=Purpose.PROPERTY,
        keys=[id],
    )

    pclass = Concept(
        name="class",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.INTEGER,
        keys=[id],
    )
    survived = Concept(
        name="survived",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.INTEGER,
        keys=[id],
    )

    survived_count = create_function_derived_concept(
        "survived_count",
        namespace,
        FunctionType.SUM,
        [survived],
        output_purpose=Purpose.METRIC,
    )

    fare = Concept(
        name="fare",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.FLOAT,
        keys=[id],
    )
    embarked = Concept(
        name="embarked",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.INTEGER,
        keys=[id],
    )
    cabin = Concept(
        name="cabin",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.STRING,
        keys=[id],
    )
    ticket = Concept(
        name="ticket",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.STRING,
        keys=[id],
    )

    last_name = Concept(
        name="last_name",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.STRING,
        keys=[id],
        lineage=Function(
            operator=FunctionType.INDEX_ACCESS,
            arguments=[
                Function(
                    operator=FunctionType.SPLIT,
                    arguments=[name, ","],
                    output_datatype=DataType.ARRAY,
                    output_purpose=Purpose.PROPERTY,
                    arg_count=2,
                ),
                1,
            ],
            output_datatype=DataType.STRING,
            output_purpose=Purpose.PROPERTY,
            arg_count=2,
        ),
    )
    all_concepts = [
        id,
        age,
        survived,
        survived_count,
        name,
        pclass,
        fare,
        cabin,
        embarked,
        ticket,
        last_name,
    ]
    for x in all_concepts:
        env.add_concept(x)

    env.add_datasource(
        Datasource(
            identifier="raw_data",
            address="raw_titanic",
            columns=[
                ColumnAssignment(alias="passengerid", concept=id),
                ColumnAssignment(alias="age", concept=age),
                ColumnAssignment(alias="survived", concept=survived),
                ColumnAssignment(alias="pclass", concept=pclass),
                ColumnAssignment(alias="name", concept=name),
                ColumnAssignment(alias="fare", concept=fare),
                ColumnAssignment(alias="cabin", concept=cabin),
                ColumnAssignment(alias="embarked", concept=embarked),
                ColumnAssignment(alias="ticket", concept=ticket),
            ],
            grain=Grain(components=[id]),
        ),
    )
    return env


def create_passenger_dimension(exec: Executor, name: str):
    exec.execute_raw_sql(f"CREATE SEQUENCE seq_{name} START 1;")
    exec.execute_raw_sql(
        f"""create table dim_{name} as 
                         SELECT passengerid id, name, age,
                          SPLIT(name, ',')[1] last_name
                            FROM raw_data

"""
    )


def create_arbitrary_dimension(exec: Executor, key: str, name: str):
    exec.execute_raw_sql(
        f"""create table dim_{name} as 
                         with tmp as 
                         (select {key}
                         from raw_data group by 1
                         )
                         SELECT  row_number() over() as id,
                         {key} as {name}
                          FROM tmp
"""
    )


def create_fact(
    exec: Executor,
    dims: Optional[list[str]] = None,
    include: Optional[list[str]] = None,
):
    exec.execute_raw_sql(
        """create table fact_titanic as 
                         SELECT 
                         row_number() OVER () as fact_key,
                         passengerid,
                         survived,
                         fare,
                         embarked,
                         b.id class_id,
                         cabin  
                         FROM raw_data a 
                         LEFT OUTER JOIN dim_class b on a.pclass=b.class
                         """
    )


def setup_normalized_engine() -> Executor:
    engine = create_engine(r"duckdb:///:memory:", future=True)
    csv = PurePath(dirname(__file__)) / "train.csv"
    df = pd.read_csv(csv)
    _ = df
    output = Executor(engine=engine, dialect=Dialects.DUCK_DB)

    output.execute_raw_sql("CREATE TABLE raw_data AS SELECT * FROM df")
    create_passenger_dimension(output, "passenger")
    create_arbitrary_dimension(output, "pclass", "class")
    create_fact(output, ["passenger"])
    return output


def setup_titanic_distributed(env: Environment):
    namespace = "passenger"
    id = Concept(
        name="id", namespace=namespace, datatype=DataType.INTEGER, purpose=Purpose.KEY
    )
    age = Concept(
        name="age",
        namespace=namespace,
        datatype=DataType.INTEGER,
        purpose=Purpose.PROPERTY,
        keys=[id],
    )

    name = Concept(
        name="name",
        namespace=namespace,
        datatype=DataType.STRING,
        purpose=Purpose.PROPERTY,
        keys=[id],
    )
    class_id = Concept(
        name="_class_id",
        namespace=namespace,
        purpose=Purpose.KEY,
        datatype=DataType.INTEGER,
        # keys=[id],
    )
    pclass = Concept(
        name="class",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.INTEGER,
        keys=[class_id],
    )
    survived = Concept(
        name="survived",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.BOOL,
        keys=[id],
    )
    survived = Concept(
        name="survived",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.BOOL,
        keys=[id],
    )

    survived_count = create_function_derived_concept(
        "survived_count", namespace, FunctionType.SUM, [survived]
    )

    fare = Concept(
        name="fare",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.FLOAT,
        keys=[id],
    )
    embarked = Concept(
        name="embarked",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.INTEGER,
        keys=[id],
    )
    cabin = Concept(
        name="cabin",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.STRING,
        keys=[id],
    )
    last_name = Concept(
        name="last_name",
        namespace=namespace,
        purpose=Purpose.PROPERTY,
        datatype=DataType.STRING,
        keys=[id],
        lineage=Function(
            operator=FunctionType.INDEX_ACCESS,
            arguments=[
                Function(
                    operator=FunctionType.SPLIT,
                    arguments=[name, ","],
                    output_datatype=DataType.ARRAY,
                    output_purpose=Purpose.PROPERTY,
                    arg_count=2,
                ),
                1,
            ],
            output_datatype=DataType.STRING,
            output_purpose=Purpose.PROPERTY,
            arg_count=2,
        ),
    )
    for x in [
        id,
        age,
        survived,
        survived_count,
        name,
        pclass,
        fare,
        cabin,
        embarked,
        last_name,
    ]:
        env.add_concept(x)

    env.add_datasource(
        Datasource(
            name="dim_passenger",
            address="dim_passenger",
            columns=[
                ColumnAssignment(alias="id", concept=id),
                ColumnAssignment(alias="age", concept=age),
                ColumnAssignment(alias="name", concept=name),
                ColumnAssignment(alias="last_name", concept=last_name),
                # ColumnAssignment(alias="pclass", concept=pclass),
                # ColumnAssignment(alias="name", concept=name),
                # ColumnAssignment(alias="fare", concept=fare),
                # ColumnAssignment(alias="cabin", concept=cabin),
                # ColumnAssignment(alias="embarked", concept=embarked),
            ],
            grain=Grain(components=[id]),
        ),
    )

    env.add_datasource(
        Datasource(
            name="fact_titanic",
            address="fact_titanic",
            columns=[
                ColumnAssignment(alias="passengerid", concept=id),
                ColumnAssignment(alias="survived", concept=survived),
                ColumnAssignment(alias="class_id", concept=class_id),
                ColumnAssignment(alias="fare", concept=fare),
                ColumnAssignment(alias="cabin", concept=cabin),
                ColumnAssignment(alias="embarked", concept=embarked),
            ],
            grain=Grain(components=[id]),
        ),
    )

    env.add_datasource(
        Datasource(
            name="dim_class",
            address="dim_class",
            columns=[
                ColumnAssignment(alias="id", concept=class_id),
                ColumnAssignment(alias="class", concept=pclass),
                # ColumnAssignment(alias="fare", concept=fare),
                # ColumnAssignment(alias="cabin", concept=cabin),
                # ColumnAssignment(alias="embarked", concept=embarked),
            ],
            grain=Grain(components=[class_id]),
        ),
    )

    return env


# if main gating for python
if __name__ == "__main__":
    executor = setup_engine()

    env = Environment()
    model = setup_titanic(env)
    for c in env.concepts:
        print(c)

    executor.environment = env
    environment = env
    logger.setLevel(DEBUG)
    logger.addHandler(StreamHandler())

    question = ("HOw many passengers survived in each cabin?",)
    processed_query = build_query(
        question,
        environment,
        debug=True,
    )

    processed_query_v2 = build_query_v2(
        question,
        environment,
        debug=True,
    )

    for q in [processed_query, processed_query_v2]:
        print("output_columns")
        for key in q.output_columns:
            print(key)

        print(executor.generator.compile_statement(q))
        results = executor.execute_query(q)
        for row in results:
            print(row)



BASE_1 = """Thought Process: You are a data analyst assistant. Your job is to identify the datasource most relevant to answering a business question. 
    You should select the minimum number of namespaces that covers information. Some namespaces will reference others; if that's the case, 
    eg sales.customer; you do not need to explicitly include the customer namespace as well when looking for customer information about store sales.

    Example: if you are asked for "orders by customer", you ONLY return the orders namespace.

    Sometimes you may need multiple sources:

    Example: If you are asked for "address of all employees, and how many orders they've placed" you would return employees and orders, because not all employees may have placed orders. 

    If the question suggests you use a specific datasource, eg "using ocean shipment data", assume that is sufficient.

    If it's very specific - eg "JUST use ocean shipment data", always assume that is sufficient.

    The output to the analyst should be a VALID JSON blob with the following keys and values followed by a stopword: <EOD>:
    - namespaces: a list of namespaces to use as strings

    So a submission argument "action_input" field should look like 
    {{
    "namespaces": ["orders", "customers"]
    }}
    
    to converse, will provide a series of action responses
    
    of the below json format. the action input may be a string or json.
    
    {{
    "action": "name",
    "action_input": "input",
    "reasoning": "optional thinking"
    }}

    You have access to the following tools:

    {tools}

    Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input). 

    To start, call the 'list_namespace' tool to see what options you have.

    Continue until you believe you have found all fields required to answer the question. Never fetch the description of a namespace more than once.

    Use as many tools as needed before producing the "Final Answer" action.

    Valid "action" values: any of {tool_names} and, for your last result "Final Answer". 
    Only return "Final Answer" when you are done with all work. Never set the action to 
    "Final Answer" before you are done. 

    Provide only ONE action per $JSON_BLOB, followed by Observation:, as show:

    Follow this format in responses:

    Question: input question to answer
    Thought: consider previous and subsequent steps
    Action:
    ```
    $JSON_BLOB
    ```
    Observation: action result
    ... (repeat Thought/Action/Observation N times)

    An example series:

    Question: Get all customers who live in US zip code 10245 and how many orders they placed in 2000

    Thought: The relevant namespace to get all customers would be the customers namespace, and orders would come from orders. Let me confirm the orders namespace description.
    Action:
    ```
    {{
        "action": "get_namespace_description",
        "action_input": "customers"
    }}
    ```
    Observation: <some description>
    Thought: Customer data, but no order data. I will now check the orders namespace to see what is available there:
    Action:
    ```
    {{
        "action": "get_namespace_description",
        "action_input": "orders"
    }}
    ```
    Observation: <some description>
    Thought: I should check my answer
    Action:
    ```
    {{
        "action": "validate_response",
        "action_input": {{
            "namespaces": [
                "orders", "customers"
            ]
            }},
        "reasoning": "I have verified I can use the customer namespace to get all info on customers, and the order namespace to get all info on orders."
        
    }}
    ```
    Observation: "looks good!"

    Once you have used any tools (listed below) as needed, you will produce your final result in this format. After producing your
    final answer, you cannot take any more steps.

    A final answer would look like this:
    ```
    {{
        "action": "Final Answer",
        "action_input": {{
            "namespaces": [
                <string namespace list>
            ]
            }},
        "reasoning": "Ready to submit!"
    }}
    ```
    You can access the the following namespaces:
    {namespaces}

    Begin! Reminder to ALWAYS respond with a valid json blob of an action. Always use tools. Respond directly if appropriate. Format is Action:```$JSON_BLOB``` Observation:"""



import json

from langchain.tools import StructuredTool, Tool
from trilogy.core.models import (
    Concept,
    Environment,
)

from trilogy_nlp.exceptions import ValidationPassedException
from trilogy_nlp.helpers import is_relevent_concept
from trilogy_nlp.instrumentation import EventTracker
from trilogy_nlp.llm_interface.validation import (
    VALID_STATUS,
    ValidateResponseInterface,
    validate_response,
)
from trilogy_nlp.tools import get_today_date


def concept_to_string(concept: Concept) -> str:
    return concept.address


def get_fields(environment: Environment, search: str, *args, **kwargs) -> str:
    return json.dumps(
        {
            "fields": [
                (
                    {
                        "name": concept_to_string(x),
                        "description": x.metadata.description
                        + f" A {x.datatype.value.lower()} field.",
                    }
                    if x.metadata.description
                    else {
                        "name": concept_to_string(x),
                        "description": f"A {x.datatype.value.lower()} field.",
                    }
                )
                for x in environment.concepts.values()
                if is_relevent_concept(x)
            ]
        }
    )


def get_help(x):
    raise SyntaxError(x)


def submit_response(environment, prompt, **kwargs):
    response, ir = validate_response(
        environment=environment,
        prompt=prompt,
        **kwargs,
    )
    if response["status"] == VALID_STATUS:
        raise ValidationPassedException(ir=ir)
    return response


def sql_agent_tools(
    environment, prompt: str, event_tracker: EventTracker | None = None
):
    def validate_response_wrapper(**kwargs):
        response, ir = validate_response(
            environment=environment,
            prompt=prompt,
            event_tracker=event_tracker,
            **kwargs,
        )
        return response

    def submit_wrapper(**kwargs):
        return submit_response(
            environment=environment,
            prompt=prompt,
            **kwargs,
        )

    tools = [
        StructuredTool(
            name="validate_response",
            description="""
            Check that a response is formatted properly and accurate before your final answer. Always call this with the complete final response before reporting a Final Answer!
            If the response is not correct, the "valid" argument will be false and it will return an array of errors. If it is correct, it will return "true" for the valid argument.
            You must fix all errors returned by validation before submitting. If you are not sure what the error means, ask for help using the get_validation_help tool.
            """,
            func=validate_response_wrapper,
            args_schema=ValidateResponseInterface,
            handle_tool_error=True,
        ),
        StructuredTool(
            name="submit_answer",
            description="""
            Submit your final answer. It will be validated, so this function may return errors to correct, similar to the validate_response function. If it passes, the answer will be submitted.
            """,
            func=submit_wrapper,
            args_schema=ValidateResponseInterface,
        ),
        Tool.from_function(
            func=lambda x: get_fields(environment, x),
            name="get_fields",
            description="""
            Returns array of json objects containing the names of actual fields that can be referenced, with a description if it exists. Fields always need to be referenced by exact name. Queries operate on fields only, by exact name.
            """,
        ),
        Tool.from_function(
            func=get_today_date,
            name="get_today_date",
            description="""
            Use to get the date of today.
            """,
        ),
    ]
    return tools



from trilogy.core.enums import Purpose
from trilogy.core.models import Concept, DataType, Environment

from trilogy_nlp.main import parse_query


def test_e2e_basic(engine):
    # grab the model we want to parse
    environment = Environment()
    id = Concept(
        name="id", namespace="user", datatype=DataType.INTEGER, purpose=Purpose.KEY
    )
    location = Concept(
        name="location",
        namespace="user",
        datatype=DataType.STRING,
        purpose=Purpose.PROPERTY,
        keys={id.address},
    )
    name = Concept(
        name="name",
        namespace="user",
        datatype=DataType.STRING,
        purpose=Purpose.PROPERTY,
        keys={id.address},
    )

    environment.add_concept(concept=id)
    environment.add_concept(concept=location)
    environment.add_concept(concept=name)

    location = environment.concepts["user.location"]

    processed_query = parse_query(
        input_text="Which users are in germany?",
        input_environment=environment,
        llm=engine,
    )
    assert location in [x for x in processed_query.where_clause.concept_arguments]



BASE_1 = """You are a data analyst assistant. Your job is to turn unstructured business questions into structured queries against a database with a known schema.

    Your goal will be to create a final output in a JSON spec defined below. Do your best to get to the most complete answer possible using all tools. 

    OUTPUT STRUCTURE:
    The key structure in your output will be a Column, a recursive json structure containing a name and an optional calculation sub-structure.  

    If the Column does not have a calculation, the name must reference a name provided in the database already or previously defined by a Column object.

    A Column Object is json with two fields:
    -- name: the field being referenced or a new derived name created in a previous Column object with a calculation. If there is a calculation, this should always be a new derived name you came up with. That name must be unique; a calculation cannot reference an input with the same name as the output concept.
    -- calculation: An optional calculation object. Only include a calculation if you need to create a new column because there is not a good match from the existing field list. 

    If the user requests something that would require two levels of aggregation to express in a language such as SQL - like an "average" of a "sum" - 
    use nested calculations or references to previously defined columns to express the concept. Ensure
    each level of calculation uses the "over" clause to define the level to group to.

    Reinforcement: make sure you use the "over" clause on any aggregate calculation if you need to group to a particular level. In general, you'll use this for any aggregate function w/ the output columns you select, except
    if the query specifically asks for an aggregate at a different level. 

    IMPORTANT: If you want an aggregate to be over all the dimensions you return, put "*" in the over clause.

    Example:

        "calculation": {{
                "operator": "SUM",
                "arguments": [
                    {{
                        "name": "store.return_value"
                    }}
                ]
                "over": [
                 "*"
                ]
            }}

    For example, when calculating a ratio of costs to revenue by state, you would have a nested calculation of costs by state and a calculation of revenue by state, and a parent division calculation with no over clause.

    For example, to get the average customer revenue by store, you would first sum the revenue by customer, then average that sum by store.

    Examples:
    # basic column
    # use this whenever you just need to return a value that fits the question. No calculation is required.
            {{
                "name": "store.id"
            }}
        or
            {{
                "name": "customer.address.zip"
            }}

    # column with calculation over another column
            {{
                "name": "total_per_customer_store_returns",
                "calculation": {{
                    "operator": "SUM",
                    "arguments": [
                        {{
                            "name": "store.return_value"
                        }}
                    ]
                    "over": [
                        {{"name": "store.id"}},
                        {{"name": "customer.id"}}
                    ]
                }}
            }}
    # column with a calculation off the previous definition, do a different granularity
            {{
                "name": "average_return_by_store_state",
                "calculation": {{
                    "operator": "AVG",
                    "arguments": [
                        {{
                            "name": "total_per_customer_store_returns"
                        }}
                    ],
                    "over": [
                    
                        {{"name": "store.state.id"}}
                    ]
                }}
            }}


    A Literal Object is json with these fields:
    -- value: the literal value ('1', 'abc',  1.0, etc), expressed as a string, or a Calculation Object
    -- type: the type of the value ('float', 'string', 'int', 'bool'), expressed as a string

    Examples: 
    # with constant
        {{"value": "1.2", "type": "float"}},
    # with null 
        {{"value": "null", "type": "null"}},
    # with calculation
        {{"value": {{
                        "operator": "MULTIPLY",
                        "arguments": [
                            {{"value": "1.2", "type": "float"}},
                            {{"name": "average_return_by_store"}}
                        ]
                    }},
            "type" : "float"
        }}

    A Calculation Object is json with three fields:
    -- operator: a function to call with those arguments. [SUM, AVG, COUNT, MAX, MIN, SUBSTRING, etc], expressed as a string. A calculation object MUST have an operator. This cannot be a comparison operator.
    -- arguments: a list of Column or Literal objects. If there is an operator, there MUST be arguments
    -- over: an optional list of Column objects used when an aggregate calculation needs to group over other columns (sum of revenue by state and county, for example). Mandatoryfor aggregations.

    A Comparison object is JSON with three fields:
    -- operator: the comparison operator, one of "=", "in", "<", ">", "<=", "like", or ">=". Use two comparisons to represent a between
    -- left: A Column or Literal object
    -- right: A Column or Literal object

    A ConditionGroup object is JSON with two fields used to create boolean filtering constructs. You can nest ConditionGroups to create complex filtering conditions.
    -- values: a list if Comparison Objects or ConditionGroups
    -- boolean: 'and' or 'or' (lowercase, no quotes)

    All together, the input for validation and final submission should be a VALID JSON blob with the following keys and values followed by a stopword: <EOD>:
    - output_columns: a list of columns to return to the user as Column objects
    - limit: a number of records to limit the results to, -1 if none specified
    - order: a list of columns to order the results by, with the option to specify ascending or descending
        -- column_name: a column name to order by; must reference value in columns
        -- order: the direction of ordering, "asc", "desc", "asc nulls first", "asc nulls last", "desc nulls first", "desc nulls last". Only specify null order when the prompt requests it.
    - filtering: an object with a single argument
        -- root: a ConditionGroup object, containing all conditions + columns used for filtering the results

    You have access to the following tools:

    {tools}

    Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input). 

    You will get essential information from using tools before submitting your answer.

    Use as many tools as needed, and always validate, before producing the "submit_answer" action.

    Valid "action" values: any of {tool_names} and, for your last result "submit_answer". 

    Only call the "submit_answer" tol when you are done with all work. Never set the action to "submit_answer" before you are done, and never set the action to submit_answer without some columns returned.

    You should always call the the validate_response tool with your candidate answer before submitting it as the final answer.

    Provide only ONE action per $JSON_BLOB, followed by Observation:, as show:

    <start example>
    ```
    {{
        "action": $TOOL_NAME,
        "action_input": $INPUT,
        "reasoning": "Your thinking"
    }}
    ```
    Observation:
    <end example>

    Action input is in JSON format, not as a JSON string blob (No escaping!)

    Follow this format in responses:

    Question: input question to answer
    Thought: consider previous and subsequent steps
    Action: 
    ```
    $JSON_BLOB
    ```
    Observation: <action result>
    ... (repeat Thought/Action/Result N times)
    Action: 
    ```
    {{
        "action": "submit_answer",
        "action_input": <action_input>,
        "reasoning": "Your thinking"
    }}
    ```
    Observation: <action result>
    An example series:

    Question: Get the total revenue dollars by order and customer id for stores in the zip code 1025 in the year 2000 where the total sales price of the items in the order was more than 100 dollars and the total revenue of the order was more than 10 dollars?
    Thought: I should start by using any available tools to inform my action.
    Action:
    ```
    {{
        "action": "get_database_description",
        "action_input": ""
        "reasoning": "I should get the available fields in the database."
    }}
    Observation: {{"fields": [<a list if fields in format {{"name": "field_name", <optional description>]}} }}
    Action:
    ```
    {{
        "action": "validate_response",
        "action_input": {{
            "output_columns": [
                {{"name": "store.order.id"}},
                {{"name": "store.order.customer.id"}},
                {{"name": "revenue_sum", 
                    "calculation": {{
                        "operator":"SUM", 
                        "arguments": [
                                {{
                                "name": "revenue_dollars",
                                "calculation" : {{
                                    "operator": "MULTIPLY",
                                    "arguments": [
                                        {{ "name": "store.order.revenue_cents" }}
                                        ]
                                    }}
                                }}
                            ],
                            "over": [
                                {{"name": "store.order.id"}},
                                {{"name": "store.order.customer.id"}}]
                        }}
                }}
            ],
            "filtering": {{
                "root": {{
                    "values": [
                    {{
                        "operator": "="
                        "left": {{"name": "store.zip_code"}},
                        "right": {{"value":"10245", "type":"integer"}},
                        
                    }},
                    {{
                        "operator": "="
                        "left": {{"name": "store.order.date.year" }},
                        "right": {{"value":"2000", "type":"integer"}},
                        
                    }},
                    {{
                        "operator": ">"
                        "left": {{"name": "revenue_sum" }},
                        "right": {{"value":"10", "type":"float"}},
                        
                    }},
                    {{  
                        "operator": ">"
                        "left": {{
                            "name": "sales_price_sum_by_store", 
                            "calculation": {{"operator":"SUM", 
                                "arguments": [
                                    {{ "name": "item.sales_price"}}
                                    ],
                                "over": [
                                    {{ "name": "store.order.id"}}, 
                                    {{ "name": "store.id"}}
                                ]
                                }}
                            }},
                        "right": {{"value":"100.0", "type":"float"}},
                        
                        
                    }}
                    ],
                    "boolean": "and"
                    }},
            }}
            "order": [
                {{"column_name": "customer_id", "order": "asc nulls first"}},
                {{"column_name": "revenue_sum", "order": "desc"}}
            ],
            "limit": 100
            }}, 
        "reasoning": "I can return order id, customer id, and the total order revenue. Order Id and customer Id are scalar values, while the total order revenue will require a calculation. I can filter to the zip code and the year, and then restrict to where the sales price over the store and order id is more than 100, which will require a calculation. Before submitting my answer, I need to validate my answer."
    }}

    IMPORTANT:
    Only include a column in the select clause if it is necessary for the final output.

    Nested Column objects with calculations can create complex derivations. This can be useful for filtering. Use nested calculations to create
    complex filtering.

    IMPORTANT: don't trust that the answer formatted a literal for filtering appropriately. For example, if the prompt asks for 'the first month of the year', you may need to filter to
    1, January, or Jan. Field descriptions will contain formatting hints that can be used for this. 
    To filter where something is not null, compare a field using "is not" as the operator to a literal of value "null" and type "null";

    example, in a comparison:
    {{
    "operator" : "is not",
    "left": {{"name": "store_sales.customer.id"}},
    "right": {{"value": "null", "type": "null"}}
    }}

    Filtering can also leverage calculations - for example, to create a filter condition for "countries with an average monthly rainfall of 2x the average on their continent", 
    the filtering clause might look like.

    "filtering": {{
            "root": {{
                "values": 
                    [
                    {{
                    "operator": ">",
                    "left": {{
                        "name": "avg_monthly_rainfall",
                        "calculation": {{
                            "operator": "AVG",
                            "arguments": [{{"name": "country.monthly_rainfall"}}],
                            "over": [
                                {{"name": "country.name"}},
                                {{"name": "date.rainfall"}}
                                ]
                            }}
                        }},
                    "right":  {{
                        "name": "continent_avg_monthly_rainfall_2x",
                        "calculation": {{
                            "operator": "MULTIPLY",
                            "arguments": [
                                    {{  "value":"2", "type":"integer"}}, 
                                    {{  "name": "continent_avg_monthly_rainfall",
                                        "calculation" : {{
                                            "operator": "AVG",
                                            "arguments": ["country.monthly_rainfall"],
                                            "over": [
                                                {{"name": "country.continent"}},
                                                {{"name": "date.rainfall"}}
                                            ]
                                        }}
                                    }}
                                    ],

                                }}
                        }}
                    }},
                    {{
                        "boolean": "or",
                        "values": [
                            {{
                                "operator": "=",
                                "left": {{"name": "country.name"}},
                                "right": {{"value": "Aruba", "type": "string"}}
                            }},
                            {{
                                "operator": "=",
                                "left": {{"name": "country.name"}},
                                "right": {{"value": "Brazil", "type": "string"}}
                            }}
                        ]
                    }}
                    ],
                "boolean": "and"
        }}
    }}

    Once you have used any tools (listed below) as needed, you will produce your final result in this format. If your final answer is wrong,
    you'll receive the prompt again with a hint that you got it wrong. You can see the output of your last three actions only.

    A final response could look like this:

    Thought: I have my final, validated answer!
    Action:
    ```
    {{
        "action": "submit_answer",
        "action_input": <VALID_JSON_WITH_SPEC_DEFINED_ABOVE>,
        "reasoning": "<description of your logic>"
    }}
    ```
    Observation: <applause>

    
    Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Action:```$JSON_BLOB```then Observation.

    """



from difflib import get_close_matches
from pathlib import Path

from langchain.agents import AgentExecutor, create_structured_chat_agent
from langchain.tools import StructuredTool, Tool
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from pydantic import BaseModel
from trilogy import Environment
from trilogy.core.enums import Modifier, Purpose
from trilogy.core.exceptions import InvalidSyntaxException

from trilogy_nlp.constants import logger
from trilogy_nlp.helpers import is_relevent_concept
from trilogy_nlp.instrumentation import EventTracker
from trilogy_nlp.prompts_v2.query_environment import BASE_1


class AddImportResponse(BaseModel):
    """result of initial parse request"""

    namespaces: list[str]
    reasoning: str | None = None


def get_environment_possible_imports(env: Environment) -> str:
    raw = Path(env.working_path).glob("*.preql")
    return "Observation: " + str([x.stem for x in raw if "query" not in x.stem])


def get_environment_detailed_values(env: Environment, input: str):
    new = Environment(working_path=env.working_path)
    try:
        new.parse(f"import {input} as {input};")

    except (ImportError, InvalidSyntaxException):
        suggestions = get_close_matches(input, get_environment_possible_imports(env))
        base = "Observation: This is an invalid database. Refer back to the list from the list_databases tool. Names must match exactly."
        if suggestions:
            base += f" Did you mean {suggestions[0]}?"
        return base
    # TODO: does including a description work better?
    return "Observation:" + str(
        {
            k.split(".", 1)[1]
            for k, v in new.concepts.items()
            # skipp hidden values
            if is_relevent_concept(v)
        }
    )


def validate_response(
    namespaces: list[str],
    environment: Environment,
    reasoning: str | None = None,
    event_tracker: EventTracker | None = None,
):
    event_tracker = event_tracker or EventTracker()
    possible = get_environment_possible_imports(environment)
    if not all(x in possible for x in namespaces):
        event_tracker.count(event_tracker.etype.ENVIRONMENT_VALIDATION_FAILED)
        return {
            "status": "invalid",
            "error": f"Not all of those namespaces exist! You must pick from {possible}",
        }
    event_tracker.count(event_tracker.etype.ENVIRONMENT_VALIDATION_PASSED)
    return {
        "status": "valid",
    }


def environment_agent_tools(environment, event_tracker: EventTracker | None = None):
    def get_import_wrapper(*args, **kwargs):
        return get_environment_possible_imports(environment)

    def validate_response_wrapper(namespaces: list[str], reasoning: str | None = None):
        return validate_response(
            namespaces=namespaces,
            reasoning=reasoning,
            environment=environment,
            event_tracker=event_tracker,
        )

    tools = [
        # Tool.from_function(
        #     func=get_import_wrapper,
        #     name="list_databases",
        #     description="""
        #     Describe the databases you can look at. Takes empty argument string.""",
        #     handle_tool_error='Argument is an EMPTY STRING, rendered as "". {} is not valid. Do not call with {} ',
        # ),
        Tool.from_function(
            func=lambda x: get_environment_detailed_values(environment, x),
            name="get_namespace_description",
            description="""
           Describe the namespace and general groupings of fields available. Call with a namespace name.""",
            handle_tool_error=True,
        ),
        StructuredTool(
            name="validate_response",
            description="""
            Check that a response is formatted properly and accurate before your final answer. Always call this!
            """,
            func=validate_response_wrapper,
            args_schema=AddImportResponse,
        ),
    ]
    return tools


def llm_loop(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    event_tracker: EventTracker | None = None,
) -> AddImportResponse:
    system = BASE_1
    human = """{input}

    {agent_scratchpad}

    (reminder to respond in a JSON blob no matter what)"""

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", system),
            MessagesPlaceholder("chat_history", optional=True),
            ("human", human),
        ]
    )
    prompt = prompt.partial(
        namespaces=get_environment_possible_imports(input_environment)
    )

    tools = environment_agent_tools(input_environment, event_tracker=event_tracker)
    chat_agent = create_structured_chat_agent(
        llm=llm,
        tools=tools,
        prompt=prompt,
    )
    agent_executor = AgentExecutor(
        agent=chat_agent, tools=tools, verbose=debug, handle_parsing_errors=True  # type: ignore
    )

    result = agent_executor.invoke({"input": input_text})
    output = result["output"]
    if "Agent stopped due to iteration limit" in output:
        raise TimeoutError(output)
    if isinstance(output, str):
        return AddImportResponse.model_validate_json(output)
    elif isinstance(output, dict):
        return AddImportResponse.model_validate(output)
    else:
        raise ValueError(f"Unable to parse LLM response {type(output)} {output}")


def select_required_import(
    input_text: str,
    environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    event_tracker: EventTracker | None = None,
) -> AddImportResponse:
    exception = None
    MAX_ATTEMPTS = 3
    attempts = 0
    while attempts < MAX_ATTEMPTS:
        try:
            return llm_loop(
                input_text, environment, llm, debug=debug, event_tracker=event_tracker
            )
        except Exception as e:
            if (
                "The model produced invalid content. Consider modifying your prompt if you are seeing this error persistently"
                in str(e)
            ):
                continue
            if debug:
                raise e
            logger.error("Error in select_required_import: %s", e)
            exception = e
            attempts += 1
    if exception:
        raise exception
    raise ValueError(
        f"Unable to get parseable response after {MAX_ATTEMPTS} attempts; "
    )


def build_env_and_imports(
    input_text: str,
    working_path: Path,
    llm: BaseLanguageModel,
    debug: bool = False,
    event_tracker: EventTracker | None = None,
):
    base = Environment(working_path=working_path)
    response: AddImportResponse = select_required_import(
        input_text, base, llm=llm, debug=debug, event_tracker=event_tracker
    )

    for x in response.namespaces:
        base.parse(f"""import {x} as {x};""")

    # automatically merge concepts between the imported domains
    for k, concept in base.concepts.items():
        if not concept.namespace:
            continue
        # only merge on keys automatically
        if not concept.purpose == Purpose.KEY:
            continue
        # don't merge on date
        # TODO: be less hacky in terms of what default merges we should do
        if "date" in concept.namespace:
            continue
        if "__preql_internal" in k:
            continue
        pre_namespace_root = k.split(".", 1)[1]
        if "." not in pre_namespace_root:
            continue
        if pre_namespace_root in base.concepts:
            base.merge_concept(
                concept, base.concepts[pre_namespace_root], modifiers=[Modifier.PARTIAL]
            )

    return base



import sqlite3

from trilogy_nlp.cache_providers.base import BaseCache
from trilogy_nlp.constants import logger

DEFAULT_SQLITE_ADDRESS = "local_prompt_cache.db"


class SqlliteCache(BaseCache):
    def __init__(self, sqlite_address: str = DEFAULT_SQLITE_ADDRESS):
        self.sqlite_address = sqlite_address

    def retrieve(self, prompt_hash: str) -> str | None:
        logger.debug(f"checking for cache with prompt hash {prompt_hash}")
        con = sqlite3.connect(self.sqlite_address)
        cur = con.cursor()
        cur.execute(
            "create table if not exists prompt_cache (cache_id string, prompt_type string, response string)"
        )
        res = cur.execute(
            "select response, prompt_type from prompt_cache where cache_id = ?",
            (prompt_hash,),
        )
        current = res.fetchone()
        if current:
            logger.debug(f"Got cached response of type {current[1]}")
            return current[0]
        logger.debug("No cache available for key")
        return None

    def store(self, prompt_hash: str, category: str, result: str):
        con = sqlite3.connect(self.sqlite_address)
        cur = con.cursor()
        cur.execute(
            "create table if not exists prompt_cache (cache_id string, prompt_type string, response string)"
        )
        cur.execute(
            "insert into prompt_cache select ?, ?,  ?", (prompt_hash, category, result)
        )
        con.commit()



# patched method while waiting for upstream PR to be merged
import logging
from typing import (
    Any,
    Callable,
    Union,
)

import langchain_community.llms.openai as langchain_openai
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)


def patch_langchain():
    def create_retry_decorator(
        llm: Union[langchain_openai.BaseOpenAI, langchain_openai.OpenAIChat]
    ) -> Callable[[Any], Any]:
        import openai

        min_seconds = 4
        max_seconds = 10
        # Wait 2^x * 1 second between each retry starting with
        # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
        return retry(
            reraise=True,
            stop=stop_after_attempt(llm.max_retries),
            wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
            retry=(
                retry_if_exception_type(openai.APITimeoutError)
                | retry_if_exception_type(openai.APIError)
                | retry_if_exception_type(openai.APIConnectionError)
            ),
            before_sleep=before_sleep_log(langchain_openai.logger, logging.WARNING),
        )

    langchain_openai._create_retry_decorator = create_retry_decorator



# llm will not know how to use these; better not to include
COMPLICATED_FUNCTIONS = ["ALIAS", "CUSTOM", "GROUP", "UNNEST", "ATTR_ACCESS"]


MAGIC_GENAI_DESCRIPTION = "local to select"

DEFAULT_GENAI_NAMESPACE = "local_nlp"



# from trilogy.core.constants import
from enum import Enum
from typing import Optional, Union

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)
from trilogy.core.enums import BooleanOperator, ComparisonOperator, Ordering


class MagicEnum(Enum):
    STAR_OP = "*"


class OrderResultV2(BaseModel):
    """The result of the order prompt"""

    model_config = ConfigDict(extra="forbid")
    column_name: str = Field(
        validation_alias=AliasChoices("column_name", "column", "name")
    )
    order: Ordering


class Literal(BaseModel):
    model_config = ConfigDict(extra="forbid")
    value: Union[str, "Calculation", list[str], list[int], list[float]]
    type: str
    # we never want this to be provided, but if it exists, use it preferentially
    # calculation: Optional["Calculation"] = None


class Calculation(BaseModel):
    model_config = ConfigDict(extra="forbid")
    arguments: list[Union["Column", Literal, "Calculation"]]
    operator: str
    over: list["Column"] | None = None


class Column(BaseModel):
    model_config = ConfigDict(extra="forbid")
    name: str
    calculation: Optional[Calculation] = None

    def __hash__(self):
        return hash(self.name + str(self.calculation))


Calculation.model_rebuild()
Literal.model_rebuild()


class NLPComparisonOperator(Enum):
    BETWEEN = "between"

    @classmethod
    def _missing_(cls, value):
        if str(value) == "<>":
            return ComparisonOperator.NE
        if not isinstance(value, list) and " " in str(value):
            value = str(value).split()
        if isinstance(value, list):
            processed = [str(v).lower() for v in value]
            if processed == ["not", "in"]:
                return ComparisonOperator.NOT_IN
            if processed == ["is", "not"]:
                return ComparisonOperator.IS_NOT
            if value == ["in"]:
                return ComparisonOperator.IN
        return super()._missing_(str(value).lower())


# hacky, but merge two enums
NLPComparisonOperator._member_map_.update(ComparisonOperator._member_map_)


class NLPConditions(BaseModel):
    left: Column | Literal
    # right can be a contains operator
    right: Column | Literal | list[Column | Literal]
    operator: ComparisonOperator


class NLPComparisonGroup(BaseModel):
    boolean: BooleanOperator
    values: list[Union[NLPConditions, "NLPComparisonGroup"]]

    @field_validator("boolean", mode="before")
    @classmethod
    def check_boolean(cls, v: str, info: ValidationInfo) -> BooleanOperator:
        if isinstance(v, str):
            return BooleanOperator(v.lower())
        return BooleanOperator(v)


class FilterResultV2(BaseModel):
    """The result of the filter prompt"""

    root: NLPComparisonGroup


NLPComparisonGroup.model_rebuild()


class InitialParseResponseV2(BaseModel):
    """The result of the initial parse"""

    output_columns: list[Column]
    limit: Optional[int] = 100
    order: Optional[list[OrderResultV2]] = None
    filtering: Optional[FilterResultV2] = None

    @field_validator("filtering", mode="plain")
    @classmethod
    def filtering_validation(cls, v):
        if isinstance(v, dict):
            return FilterResultV2.model_validate(v)
        return FilterResultV2.model_validate(v)



from langchain.agents import AgentExecutor, create_structured_chat_agent
from langchain.agents.agent import OutputParserException
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from networkx import DiGraph, topological_sort
from trilogy.core.models import (
    AggregateWrapper,
    Concept,
    ConceptDeclarationStatement,
    ConceptTransform,
    Environment,
    FilterItem,
    Function,
    Grain,
    ProcessedQuery,
    SelectItem,
    SelectStatement,
    WindowItem,
)
from trilogy.core.query_processor import process_query
from trilogy.parsing.render import Renderer

from trilogy_nlp.config import DEFAULT_CONFIG
from trilogy_nlp.constants import logger
from trilogy_nlp.exceptions import ValidationPassedException
from trilogy_nlp.helpers import safe_limit
from trilogy_nlp.instrumentation import EventTracker
from trilogy_nlp.llm_interface.constants import MAGIC_GENAI_DESCRIPTION
from trilogy_nlp.llm_interface.models import Column, InitialParseResponseV2
from trilogy_nlp.llm_interface.parsing import (
    create_column,
    generate_having_and_where,
    parse_filtering,
    parse_order,
)
from trilogy_nlp.llm_interface.tools import sql_agent_tools
from trilogy_nlp.prompts_v2.query_system import BASE_1
from trilogy_nlp.tools import get_wiki_tool


def is_local_derived(x: Concept) -> bool:
    return x.metadata.description == MAGIC_GENAI_DESCRIPTION


def handle_parsing_error(x: OutputParserException):

    return """
The expected action format was below; please try again with the correct format.
Note that the action is the name of the tool you want to use, and the action_input is the input you want to provide to that tool, and must be valid JSON.

Action:
```
{{
    "action": $TOOL_NAME,
    "action_input": $INPUT,
    "reasoning": "Your thinking"
}}
```
Observation:
"""


def llm_loop(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    additional_context: str | None = None,
    debug: bool = False,
    event_tracker: EventTracker | None = None,
) -> SelectStatement:

    human = """{input}

    {agent_scratchpad}

    (reminder to respond in a JSON blob no matter what)"""

    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", BASE_1),
            MessagesPlaceholder("chat_history", optional=True),
            ("human", human),
        ]
    )

    tools = sql_agent_tools(input_environment, input_text, event_tracker) + [
        get_wiki_tool()
    ]
    chat_agent = create_structured_chat_agent(
        llm=llm,
        tools=tools,
        prompt=prompt,
    )
    agent_executor = AgentExecutor(
        agent=chat_agent,
        tools=tools,
        verbose=True if debug else False,
        max_iterations=DEFAULT_CONFIG.LLM_VALIDATION_ATTEMPTS,
        early_stopping_method="force",
        handle_parsing_errors=handle_parsing_error,
    )
    attempts = 0
    if additional_context:
        input_text += additional_context
    error = None
    while attempts < 4:
        output = None
        try:
            output = agent_executor.invoke({"input": input_text})
            return ir_to_query(
                InitialParseResponseV2.model_validate(output["output"]),
                input_environment,
                debug=debug,
            )
        except ValidationPassedException as e:
            ir = e.ir
            return ir_to_query(ir, input_environment=input_environment, debug=True)

        except Exception as e:
            error = e
            logger.error(
                f"Error in main execution llm loop with output {output}: {str(e)}"
            )
            if (
                "peer closed connection without sending complete message body (incomplete chunked read)"
                in str(e)
            ):
                attempts + 1
                continue
            # don't retry an unknown error
            break
    if error:
        raise error
    raise ValueError(f"Unable to get parseable response after {attempts} attempts")


def determine_ordering(columns: list[Column]):
    edges = []

    def handle_column(column: Column):
        calculation = column.calculation
        base_name = column.name
        if not calculation:
            return
        for arg in calculation.arguments:
            if isinstance(arg, Column) and arg.name == base_name:
                base_name = base_name + "_deriv"
        for arg in calculation.arguments:
            if not isinstance(arg, Column):
                continue
            edges.append((arg.name, base_name))
            handle_column(arg)
        for arg in calculation.over or []:
            if not isinstance(arg, Column):
                continue
            edges.append((arg.name, base_name))
            handle_column(arg)

    for ic in columns:
        handle_column(ic)
    graph: DiGraph = DiGraph(edges)
    base = list(topological_sort(graph))
    for column in columns:
        if column.name not in base:
            base.append(column.name)
    return base


def sort_by_name_list(arr: list[Column], reference: list[str]):
    order_map = {value: index for index, value in enumerate(reference)}
    return sorted(arr, key=lambda x: order_map.get(x.name, float("inf")))


def ir_to_query(
    intermediate_results: InitialParseResponseV2,
    input_environment: Environment,
    debug: bool = False,
):
    ordering = determine_ordering(intermediate_results.output_columns)
    selection = [
        create_column(x, input_environment)
        for x in sort_by_name_list(intermediate_results.output_columns, ordering)
    ]
    order = parse_order(selection, intermediate_results.order or [])

    filtering = (
        parse_filtering(intermediate_results.filtering, input_environment)
        if intermediate_results.filtering
        else None
    )

    if debug:
        print("Output Concepts")
        for c in intermediate_results.output_columns:
            print(c)
        if intermediate_results.filtering:
            print("Filtering")
            print(str(intermediate_results.filtering))
        if intermediate_results.order:
            print("Ordering")
            for o in intermediate_results.order:
                print(o)
    normalized_select = [x if isinstance(x, Concept) else x.output for x in selection]
    where, having = generate_having_and_where(
        [x.address for x in normalized_select], filtering
    )
    query = SelectStatement(
        selection=[
            (
                SelectItem(content=ConceptTransform(function=x.lineage, output=x))
                if is_local_derived(x)
                and x.lineage
                and isinstance(
                    x.lineage, (Function, FilterItem, WindowItem, AggregateWrapper)
                )
                else SelectItem(content=x)
            )
            for x in normalized_select
        ],
        limit=safe_limit(intermediate_results.limit),
        order_by=order,
        where_clause=where,
        having_clause=having,
    )

    query.grain = Grain.from_concepts(
        query.output_components, where_clause=query.where_clause
    )

    if having:

        def append_child_concepts(xes: list[Concept]):

            def get_address(z):
                if isinstance(z, Concept):
                    return z.address
                elif isinstance(z, ConceptTransform):
                    return z.output.address

            for x in xes:
                if not any(
                    x.address == get_address(item.content) for item in query.selection
                ):
                    if (
                        is_local_derived(x)
                        and x.lineage
                        and isinstance(
                            x.lineage,
                            (Function, FilterItem, WindowItem, AggregateWrapper),
                        )
                    ):
                        content = ConceptTransform(function=x.lineage, output=x)
                        query.selection.append(SelectItem(content=content))
                        if x.lineage:
                            append_child_concepts(x.lineage.concept_arguments)

        append_child_concepts(having.concept_arguments)

    for parse_pass in [1, 2]:
        if parse_pass == 1:
            grain = Grain.from_concepts(
                [x.content for x in query.selection if isinstance(x.content, Concept)],
                where_clause=query.where_clause,
            )
        if parse_pass == 2:
            grain = Grain.from_concepts(
                query.output_components, where_clause=query.where_clause
            )
        query.grain = grain
        for item in query.selection:
            # we don't know the grain of an aggregate at assignment time
            # so rebuild at this point in the tree
            # TODO: simplify
            if isinstance(item.content, ConceptTransform):
                new_concept = item.content.output.with_select_context(
                    local_concepts=query.local_concepts,
                    grain=query.grain,
                    environment=input_environment,
                )
                query.local_concepts[new_concept.address] = new_concept
                if parse_pass == 2:
                    input_environment.add_concept(new_concept, force=True)
                item.content.output = new_concept
            elif isinstance(item.content, Concept):
                # Sometimes cached values here don't have the latest info
                # but we can't just use environment, as it might not have the right grain.
                item.content = item.content.with_select_context(
                    local_concepts=query.local_concepts,
                    grain=query.grain,
                    environment=input_environment,
                )
                query.local_concepts[item.content.address] = item.content

    renderer = Renderer()
    print("RENDERED QUERY")
    print("---------")
    for new_c in input_environment.concepts.values():
        if is_local_derived(new_c):
            print(renderer.to_string(ConceptDeclarationStatement(concept=new_c)))
    print(renderer.to_string(query))
    print("---------")
    query.validate_syntax(environment=input_environment)
    return query


def parse_query(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    log_info: bool = True,
    event_tracker: EventTracker | None = None,
) -> SelectStatement:
    return llm_loop(
        input_text, input_environment, llm=llm, debug=debug, event_tracker=event_tracker
    )


def build_query(
    input_text: str,
    input_environment: Environment,
    llm: BaseLanguageModel,
    debug: bool = False,
    log_info: bool = True,
    event_tracker: EventTracker | None = None,
) -> ProcessedQuery:
    query = parse_query(
        input_text,
        input_environment,
        debug=debug,
        llm=llm,
        log_info=log_info,
        event_tracker=event_tracker,
    )
    return process_query(statement=query, environment=input_environment)



import json
from pathlib import Path

from trilogy import Environment

from trilogy_nlp.llm_interface.models import InitialParseResponseV2
from trilogy_nlp.main import ir_to_query

ORDERING_TEST_CASE = """{
    "action": "Final Answer",
    "action_input": {
        "output_columns": [
            {
                "name": "store_sales.customer.state"
            },
            {
                "name": "customer_count",
                "calculation": {
                    "operator": "COUNT",
                    "arguments": [
                        {
                            "name": "store_sales.customer.id"
                        }
                    ],
                    "over": [
                        {
                            "name": "store_sales.customer.state"
                        }
                    ]
                }
            }
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": ">=",
                        "left": {
                            "name": "customer_count"
                        },
                        "right": {
                            "value": "10",
                            "type": "int"
                        }
                    },
                    {
                        "operator": ">=",
                        "left": {
                            "name": "store_sales.item.current_price"
                        },
                        "right":    {
                "name": "sales_price_filtered",
                "calculation": {
                    "operator": "MULTIPLY",
                    "arguments": [
                                    {
                "name": "average_price_by_category",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "item.current_price"
                        }
                    ],
                    "over": [
                        {
                            "name": "item.category"
                        }
                    ]
                }
            },
                        {
                            "value": "1.2",
                            "type": "float"
                        }
                    ]
                }
            }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.year"
                        },
                        "right": {
                            "value": "2001",
                            "type": "int"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.month"
                        },
                        "right": {
                            "value": "1",
                            "type": "int"
                        }
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [
            {"column_name": "customer_count", "order": "asc"},
            {"column_name": "store_sales.customer.state", "order": "asc"}
        ],
        "limit": -1
    },
    "reasoning": "The query captures all states and their respective customer counts for customers who bought items with a sales price at least 1.2 times the average sales price of all other items in the same category during January 2001, filtering for those states with at least 10 customers."
}"""


def test_ordering_resolution():
    loaded = json.loads(ORDERING_TEST_CASE)
    validated = InitialParseResponseV2.model_validate(loaded["action_input"])
    environment = Environment(working_path=Path(__file__).parent / "tpc_ds_duckdb")
    environment.add_file_import("store_sales", "store_sales")
    environment.add_file_import("item", "item")
    environment.parse("MERGE store_sales.item.id INTO ~item.id;")
    ir_to_query(validated, input_environment=environment, debug=False)


HAVING_WHERE_SPLIT = """{
    "action": "submit_answer",
    "action_input": {
        "output_columns": [
            {
                "name": "store_sales.customer.state"
            },
            {
                "name": "customer_count",
                "calculation": {
                    "operator": "COUNT",
                    "arguments": [
                        {
                            "name": "store_sales.customer.id"
                        }
                    ]
                }
            }
            
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": ">=",
                        "left": {
                            "value": "10",
                            "type": "int"
                        },
                        "right": {
                            "name": "customer_count"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.year"
                        },
                        "right": {
                            "value": "2001",
                            "type": "int"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.month"
                        },
                        "right": {
                            "value": "1",
                            "type": "int"
                        }
                    },
                    {
                        "operator": ">",
                        "left": {
                            "name": "store_sales.item.current_price"
                        },
                        "right": {
                "name": "filtered_current_price",
                "calculation": {
                    "operator": "MULTIPLY",
                    "arguments": [
                        {
                            "value": "1.2",
                            "type": "float"
                        },
                         {
                "name": "average_price_by_category",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.item.current_price"
                        }
                    ],
                    "over": [
                        {
                            "name": "store_sales.item.category"
                        }
                    ]
                }
            }
                    ]
                }
            }
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [
            {
                "column_name": "customer_count",
                "order": "asc"
            },
            {
                "column_name": "store_sales.customer.state",
                "order": "asc"
            }
        ],
        "limit": -1
    },
    "reasoning": "I have constructed the final query to include the necessary columns, filtering conditions, and ordering as specified in the prompt. This will return the desired output for states and customer counts for those purchasing items that meet the sales price criteria."
}"""


def test_having_where_split():
    loaded = json.loads(HAVING_WHERE_SPLIT)
    validated = InitialParseResponseV2.model_validate(loaded["action_input"])
    environment = Environment(working_path=Path(__file__).parent / "tpc_ds_duckdb")
    environment.add_file_import("store_sales", "store_sales")
    environment.add_file_import("item", "item")
    environment.parse("MERGE store_sales.item.id INTO ~item.id;")
    ir = ir_to_query(validated, input_environment=environment, debug=False)
    assert environment.concepts["customer_count"].address in [
        c.address for c in ir.having_clause.concept_arguments
    ]


OR_PARENTHETICAL = """{
    "action": "submit_answer",
    "action_input": {
        "output_columns": [
            {
                "name": "store_sales.item.name"
            },
            {
                "name": "average_quantity_sold",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.quantity"
                        }
                    ]
                }
            },
            {
                "name": "average_list_price",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.list_price"
                        }
                    ]
                }
            },
            {
                "name": "average_coupon_amount",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.coupon_amt"
                        }
                    ]
                }
            },
            {
                "name": "average_sales_price",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.sales_price"
                        }
                    ]
                }
            }
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.gender"
                        },
                        "right": {
                            "value": "M",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.marital_status"
                        },
                        "right": {
                            "value": "S",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.education_status"
                        },
                        "right": {
                            "value": "College",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.year"
                        },
                        "right": {
                            "value": "2000",
                            "type": "integer"
                        }
                    },
                    {
                        "boolean": "or",
                        "values": [
                            {
                                "operator": "=",
                                "left": {
                                    "name": "store_sales.promotion.channel_event"
                                },
                                "right": {
                                    "value": "N",
                                    "type": "string"
                                }
                            },
                            {
                                "operator": "=",
                                "left": {
                                    "name": "store_sales.promotion.channel_email"
                                },
                                "right": {
                                    "value": "N",
                                    "type": "string"
                                }
                            }
                        ]
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [],
        "limit": -1
    },
    "reasoning": "The response has been validated and is now ready for submission. It includes the necessary output columns and filtering conditions based on the provided prompt."
}"""


def test_parenthetical_or():
    loaded = json.loads(OR_PARENTHETICAL)
    validated = InitialParseResponseV2.model_validate(loaded["action_input"])
    environment = Environment(working_path=Path(__file__).parent / "tpc_ds_duckdb")
    environment.add_file_import("store_sales", "store_sales")
    from trilogy.parsing.render import Renderer

    ir = ir_to_query(validated, input_environment=environment, debug=False)

    query = Renderer().to_string(ir)
    assert (
        "(store_sales.promotion.channel_event = 'N' or store_sales.promotion.channel_email = 'N')"
        in query
    ), query


AGG_GRAIN_CHECK = """{
    "action": "submit_answer",
    "action_input": {
        "output_columns": [
            {
                "name": "store_sales.item.name"
            },
            {
                "name": "average_quantity_sold",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.quantity"
                        }
                    ]
                }
            },
            {
                "name": "average_list_price",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.list_price"
                        }
                    ]
                }
            },
            {
                "name": "average_coupon_amount",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.coupon_amt"
                        }
                    ]
                }
            },
            {
                "name": "average_sales_price",
                "calculation": {
                    "operator": "AVG",
                    "arguments": [
                        {
                            "name": "store_sales.sales_price"
                        }
                    ]
                }
            }
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.gender"
                        },
                        "right": {
                            "value": "M",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.marital_status"
                        },
                        "right": {
                            "value": "S",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.customer.demographics.education_status"
                        },
                        "right": {
                            "value": "College",
                            "type": "string"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {
                            "name": "store_sales.date.date.year"
                        },
                        "right": {
                            "value": "2000",
                            "type": "integer"
                        }
                    },
                    {
                        "boolean": "or",
                        "values": [
                            {
                                "operator": "=",
                                "left": {
                                    "name": "store_sales.promotion.channel_event"
                                },
                                "right": {
                                    "value": "N",
                                    "type": "string"
                                }
                            },
                            {
                                "operator": "=",
                                "left": {
                                    "name": "store_sales.promotion.channel_email"
                                },
                                "right": {
                                    "value": "N",
                                    "type": "string"
                                }
                            }
                        ]
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [
            {
                "column_name": "store_sales.item.name",
                "order": "asc"
            }
        ],
        "limit": -1
    },
    "reasoning": "The query has been constructed to satisfy the requirements of the prompt, ensuring it calculates the averages for the specified fields for the appropriate customer demographics while filtering out those who came from event or email promotions. The results will be sorted by item name in ascending order."
}"""


def test_aggregate_grain():
    loaded = json.loads(AGG_GRAIN_CHECK)
    validated = InitialParseResponseV2.model_validate(loaded["action_input"])
    environment = Environment(working_path=Path(__file__).parent / "tpc_ds_duckdb")
    environment.add_file_import("store_sales", "store_sales")

    ir = ir_to_query(validated, input_environment=environment, debug=False)
    print(ir.local_concepts["average_quantity_sold"])
    print(ir.local_concepts["average_quantity_sold"].purpose)
    assert ir.local_concepts["average_quantity_sold"].grain.components == {
        "store_sales.item.name"
    }
    for x in ir.output_components:
        if x.address == "store_sales.item.name":
            continue
        assert (
            x.grain.components == ir.grain.components
        ), f"{x.address} HAS {x.grain.components} vs {ir.grain.components}"



from enum import Enum, auto


class Provider(Enum):
    OPENAI = "openai"
    OCTO_AI = "octoai"
    LOCAL = "local"
    LLAMAFILE = "llamafile"
    GOOGLE = "google"


class CacheType(Enum):
    MEMORY = "memory"
    SQLLITE = "sqlite"


class EventType(Enum):
    OPEN_DATABASE = auto()

    ## environment
    ENVIRONMENT_VALIDATION_FAILED = auto()
    ENVIRONMENT_VALIDATION_PASSED = auto()
    ## validation
    INITIAL_VALIDATION_PARSING_FAILED = auto()

    STRING_FIELD_WITH_AGGREGATE = auto()

    OVER_CLAUSE_WITHOUT_AGGREGATE = auto()

    INVALID_FUNCTION = auto()

    INVALID_COLUMN_NAME_NO_CALCULATION = auto()

    CALCULATION_WITH_PREDEFINED_FIELD = auto()

    VALIDATION_ERROR = auto()

    ORDER_BY_NOT_SELECTED = auto()

    QUERY_VALIDATION_PASSED = auto()



class SemanticExecutionException(BaseException):
    pass


class ValidationPassedException(BaseException):
    def __init__(self, ir, *args, **kwargs):
        self.ir = ir
        super().__init__(*args, **kwargs)



import difflib
import json
from collections import defaultdict
from enum import Enum
from typing import List, Optional

from langchain_core.tools import ToolException
from pydantic import BaseModel
from pydantic_core import ErrorDetails, ValidationError

# from trilogy.core.constants import
from trilogy.core.enums import (
    ComparisonOperator,
    FunctionType,
)
from trilogy.core.exceptions import UndefinedConceptException
from trilogy.core.models import (
    DataType,
    Environment,
)

from trilogy_nlp.constants import logger
from trilogy_nlp.helpers import is_relevent_concept
from trilogy_nlp.instrumentation import EventTracker
from trilogy_nlp.llm_interface.constants import COMPLICATED_FUNCTIONS
from trilogy_nlp.llm_interface.models import (
    Calculation,
    Column,
    FilterResultV2,
    InitialParseResponseV2,
    Literal,
    NLPConditions,
    OrderResultV2,
)

VALID_STATUS = "valid"


def is_valid_function(name: str):
    return name.lower() in [item.value for item in FunctionType]


def invalid_operator_message(operator: str) -> str | None:
    try:
        ComparisonOperator(operator)
    except Exception as e:
        return str(e)
    return None


VALIDATION_CACHE: dict[str, InitialParseResponseV2] = {}


class QueryContext(Enum):
    SELECT = "COLUMN"
    FILTER = "FILTER"
    ORDER = "ORDER"

    def __str__(self) -> str:
        if self == QueryContext.SELECT:
            return "output column declaration"
        return f"{self.value.lower()} declaration"


def validate_query(
    query: dict,
    environment: Environment,
    prompt: str,
    event_tracker: EventTracker | None = None,
) -> tuple[dict, InitialParseResponseV2 | None]:
    tracker = event_tracker or EventTracker()
    try:
        parsed = InitialParseResponseV2.model_validate(query)
    except ValidationError as e:
        tracker.count(tracker.etype.INITIAL_VALIDATION_PARSING_FAILED)
        return {"status": "invalid", "error": validation_error_to_string(e)}, None
    errors = []
    # assume to start that all our select calculations are valid
    select = {col.name for col in parsed.output_columns if col.calculation}
    top_select = {col.name for col in parsed.output_columns}
    filtered_on = set()

    def validate_calculation(calc: Calculation, context: QueryContext) -> bool:
        valid = True
        if not is_valid_function(calc.operator):
            errors.append(
                f"{calc} in {context} does not use a valid function (is using {calc.operator}); check that you are using ONLY a valid option from this list: {sorted([x for x in FunctionType.__members__.keys() if x not in COMPLICATED_FUNCTIONS]) }. If the column requires no transformation, drop the calculation field.",
            )
            valid = False
            tracker.count(tracker.etype.INVALID_FUNCTION)
        if calc.over:
            for x in calc.over:
                local = validate_column(x, context)
                valid = valid and local
            if calc.operator not in ["AVG", "SUM", "MIN", "MAX", "COUNT"]:
                errors.append(
                    f"{calc} in {context} should only use 'over' when using an aggregate function [AVG, SUM, MIN, MAX, COUNT] (is using {calc.operator}); remove this 'over' (make sure it's included in any parent aggregate if needed)",
                )
                tracker.count(tracker.etype.OVER_CLAUSE_WITHOUT_AGGREGATE)
                valid = False

        for arg in calc.arguments:
            if isinstance(arg, Column):
                if arg.name in environment.concepts:
                    dtype = environment.concepts[arg.name].datatype
                    if dtype == DataType.STRING:
                        errors.append(
                            f"Aggregate function {calc.operator} in {context} is being used on a string field {arg.name}; if you need this field in the output, just return it without a calculation. (Don't forget to set the full name {arg.name} when you drop the calculation!)",
                        )
                        valid = False
                        tracker.count(tracker.etype.STRING_FIELD_WITH_AGGREGATE)
            if isinstance(arg, Column):
                local = validate_column(arg, context)
                valid = valid and local
        return valid

    def validate_literal(lit: Literal, context: QueryContext) -> bool:
        if isinstance(lit.value, str) and lit.value in environment.concepts:
            return True
        elif isinstance(lit.value, Calculation):
            return validate_calculation(lit.value, context)
        return True

    def validate_column(col: Column, context: QueryContext) -> bool:
        valid = False
        if (
            col.name not in select
            and col.name not in environment.concepts
            and not col.calculation
        ):
            recommendations = None
            tracker.count(tracker.etype.INVALID_COLUMN_NAME_NO_CALCULATION)
            try:
                environment.concepts[col.name]
            except UndefinedConceptException as e:
                recommendations = e.suggestions
            if context == QueryContext.FILTER:
                hint = "add another nested column with a calculation to define it in this filter"
            else:
                hint = "add an output column to calculate it"
            if recommendations:
                errors.append(
                    f"'{col.name}' in {context} is not a valid preexisting field returned by the get fields tool or previously defined by you; if this is a new calculated column you need to answer the question, {hint}. Did you want to use one of these? {recommendations}?",
                )
            else:
                errors.append(
                    f"'{col.name}' in {context} is not a valid preexisting field returned by the get fields tool or previously defined by you. If this is a new calculated column you need to answer the question, {hint}. If you just misspelled an existing field, maybe it should be one of {difflib.get_close_matches(col.name, [k for k,v in environment.concepts.items() if is_relevent_concept(v)], 3, 0.4)}",
                )
        elif col.name in environment.concepts:
            valid = True
            if col.calculation:
                tracker.count(tracker.etype.CALCULATION_WITH_PREDEFINED_FIELD)
                errors.append(
                    f"{col.name} in {context} is a predefined field and should not have a calculation; check that you are using the field as is, without any additional transformations.",
                )
                valid = False
        elif col.calculation:
            local = validate_calculation(col.calculation, context)
            valid = valid and local
        if valid:
            if context == QueryContext.SELECT:
                select.add(col.name)
            elif context == QueryContext.FILTER:
                filtered_on.add(col.name)
        return valid

    for x in parsed.output_columns:
        validate_column(x, QueryContext.SELECT)

    if parsed.order:
        for y in parsed.order:
            if y.column_name not in top_select:
                tracker.count(tracker.etype.ORDER_BY_NOT_SELECTED)
                errors.append(
                    f"{y.column_name} being ordered by is not in output_columns, add if needed.",
                )
    if parsed.filtering:
        root = parsed.filtering.root
        for val in root.values:
            if isinstance(val, Column):
                validate_column(val, QueryContext.FILTER)

            elif isinstance(val, NLPConditions):
                for subval in [val.left, val.right]:
                    if isinstance(subval, Column):
                        validate_column(subval, QueryContext.FILTER)
                    elif isinstance(subval, Literal):
                        validate_literal(subval, QueryContext.FILTER)
                if isinstance(val.left, Column) and isinstance(val.right, Column):
                    if val.left.name == val.right.name:
                        errors.append(
                            f"Comparison {val} has the same column name on both sides; this is a meaningless comparison. Check that you are comparing two different fields.",
                        )

    if errors:
        return {
            "status": "invalid",
            "errors": {f"Error {idx+1}: {error}" for idx, error in enumerate(errors)},
        }, parsed
    tracker.count(tracker.etype.QUERY_VALIDATION_PASSED)
    tips = [
        f'Congrats! No validation errors - looking good! A few final checks - if you believe these do not apply, go ahead and submit this response as your final answer! (or make a tweak or two!) First, check that you have all fields you think are required in the output. Even better; keep it clean - if an output column is added only to be used in filtering and not explicitly asked for, you might be able to move it out of the select (move the entire definition into a nested reference).  The original prompt was: "{prompt}"!'
    ]
    for all_address in select.union(filtered_on):
        if all_address in environment.concepts:
            concept = environment.concepts[all_address]
            if concept.metadata.description:
                tips.append(
                    f'And another one - for the field {all_address}, reminder that the field description is "{concept.metadata.description}". Double check that any filtering on this field matches the described format! (this is a reminder, not an error)'
                )
    VALIDATION_CACHE[prompt] = parsed
    return {"status": VALID_STATUS, "tips": tips}, parsed


def validation_error_to_string(err: ValidationError):
    # inject in new context on failed answer
    raw_error = str(err)
    errors = err.errors()
    logger.error(f"Validation error: {errors}")
    if "filtering.root." in raw_error:
        missing = []
        path_freq: dict[str, int] = defaultdict(lambda: 0)
        path_map: dict[str, list[str]] = defaultdict(lambda: [])
        for validation_error in errors:
            e: ErrorDetails = validation_error
            if e["type"] == "missing":
                path = "".join([str(x) for x in e["loc"][:-1]])
                path_freq[path] += 1
                key = e["loc"][-1]
                message = f'Missing "{e["loc"][-1]}" in this JSON object you provided: {json.dumps(e["input"], indent=4)}. You may also be using the wrong object. Double check Literal and Column formats.'
                if key == "type" and "value" in e["input"]:
                    nested = e["input"]["value"]
                    if "type" in nested:
                        message = f'Missing "{e["loc"][-1]}". You have have overly nested a literal in {json.dumps(e["input"], indent=4)}. Flatten the inner object {json.dumps(nested, indent=4)} out into the parent.'
                path_map[path].append(message)
                missing.append(message)
            elif e["type"] == "extra_forbidden":
                path = "".join([str(x) for x in e["loc"][:-1]])
                path_freq[path] += 1
                message = f'Extra invalid key "{e["loc"][-1]}" in this JSON object you provided: {json.dumps(e["input"], indent=4)}'
                path_map[path].append(message)
                missing.append(message)
            else:
                missing_path = ".".join(
                    [
                        str(v)
                        for v in e["loc"]
                        if str(v)
                        not in [
                            "NLPComparisonGroup",
                            "NLPConditions",
                            "Literal",
                            "Column",
                        ]
                    ]
                )
                missing.append(missing_path)
        min_path = min(path_freq, key=path_freq.get)  # type: ignore
        locations = path_map[min_path]
        raw_error = f"Syntax error in your filtering clause. Check that you only use valid keys for Literal, Comparison, and Column objects, and have all required keys. Look at these locations: {locations}"
    else:
        for e in errors:
            if e["type"] == "missing":
                raw_error = f'Could not parse {e["input"]}. Maybe missing {e["loc"][-1]}? You might have incorrect JSON formatting or the key is actually missing. Double check Literal and Column formats.'
    return raw_error


def validate_response(
    environment: Environment,
    prompt: str,
    output_columns: List[Column],
    filtering: Optional[FilterResultV2] = None,
    order: Optional[list[OrderResultV2]] = None,
    limit: int | None = None,
    event_tracker: EventTracker | None = None,
) -> tuple[dict, InitialParseResponseV2 | None]:

    base: dict[str, list[Column] | FilterResultV2 | list[OrderResultV2] | int] = {
        "output_columns": output_columns
    }
    if not output_columns:
        raise ToolException(
            "A request to validate should include at least one column. The input should be the exactly same as the input for your final answer. Generate your answer, call this tool again, and then submit it if it passes."
        )
    if filtering:
        base["filtering"] = filtering
    if order:
        base["order"] = order
    if limit:
        base["limit"] = limit
    return validate_query(
        base,
        environment=environment,
        prompt=prompt,
        event_tracker=event_tracker,
    )


class ValidateResponseInterface(BaseModel):
    # deliberately permissive interface
    # so that we can handle the error inside the tool
    output_columns: List[dict]
    filtering: Optional[dict] = None
    order: Optional[list[dict]] = None
    limit: int | None = None



from .monkeypatch import patch_langchain

patch_langchain()

from .core import NLPEngine  # noqa: E402
from .enums import CacheType, Provider  # noqa: E402
from .main import build_query  # noqa: E402

__version__ = "0.1.5"

__all__ = ["build_query", "Provider", "NLPEngine", "CacheType"]



from collections import Counter

from trilogy_nlp.enums import EventType


class EventTracker:
    etype = EventType

    def __init__(self):
        self.events: Counter[EventType] = Counter()

    def count(self, event_type: EventType):
        self.events[event_type] += 1



import json
from pathlib import Path

from trilogy import Environment

# from trilogy.dialect.base
from trilogy_nlp.llm_interface.validation import (
    Column,
    FilterResultV2,
    validate_response,
)

INVALID_FUNCTION = {
    "output_columns": [
        {"name": "store_sales.store.state"},
        {
            "name": "customer_count",
            "calculation": {
                "operator": "COUNT",
                "arguments": [{"name": "store_sales.customer.id"}],
            },
        },
    ],
    "filtering": {
        "root": {
            "values": [
                {
                    "left": {
                        "name": "store_sales.sales_price",
                        "calculation": {
                            "operator": ">=",
                            "arguments": [
                                {
                                    "name": "average_category_sales_price",
                                    "calculation": {
                                        "operator": "MULTIPLY",
                                        "arguments": [
                                            {"value": "1.2", "type": "float"},
                                            {
                                                "name": "avg_sales_price_by_category",
                                                "calculation": {
                                                    "operator": "AVG",
                                                    "arguments": [
                                                        {"name": "item.list_price"}
                                                    ],
                                                    "over": [{"name": "item.category"}],
                                                },
                                            },
                                        ],
                                    },
                                }
                            ],
                        },
                    },
                    "right": {"value": "2001-01-01", "type": "string"},
                    "operator": "like",
                }
            ],
            "boolean": "and",
        }
    },
    "order": [
        {"column_name": "customer_count", "order": "asc"},
        {"column_name": "store_sales.store.state", "order": "asc"},
    ],
    "limit": -1,
}


def test_validate_response_invalid_function():
    # check to make sure that the invalid function is detected
    filtering = FilterResultV2.model_validate(INVALID_FUNCTION["filtering"])
    columns = [Column.model_validate(x) for x in INVALID_FUNCTION["output_columns"]]
    env = Environment()
    env.parse(
        """
key store_sales.store.id int;
key store_sales.customer.id int;
key store_sales.store.state string;"""
    )
    response, ir = validate_response(
        environment=env,
        output_columns=columns,
        filtering=filtering,
        order=None,
        limit=100,
        prompt="shenanigans",
    )

    assert response["status"] == "invalid", response
    errors = response["errors"]
    assert "does not use a valid function" in str(errors), errors


TEST_INVALID_FUNCTION_TWO = {
    "output_columns": [
        {"name": "store_sales.customer.state"},
        {
            "name": "customer_count",
            "calculation": {
                "operator": "COUNT",
                "arguments": [{"name": "store_sales.customer.id"}],
            },
        },
    ],
    "filtering": {
        "root": {
            "values": [
                {
                    "left": {
                        "name": "store_sales.item.current_price",
                        "calculation": {
                            "operator": "GREATER_THAN_OR_EQUAL",
                            "arguments": [
                                {
                                    "name": "average_sales_price_in_category",
                                    "calculation": {
                                        "operator": "MULTIPLY",
                                        "arguments": [
                                            {"value": "1.2", "type": "float"},
                                            {
                                                "name": "average_sales_price_in_category",
                                                "calculation": {
                                                    "operator": "AVG",
                                                    "arguments": [
                                                        {
                                                            "name": "store_sales.item.current_price"
                                                        }
                                                    ],
                                                },
                                            },
                                        ],
                                    },
                                }
                            ],
                        },
                    },
                    "right": {"value": "0", "type": "int"},
                    "operator": ">",
                },
                {
                    "left": {"name": "store_sales.date.date.year"},
                    "right": {"value": "2001", "type": "int"},
                    "operator": "=",
                },
                {
                    "left": {"name": "store_sales.date.date.month"},
                    "right": {"value": "1", "type": "int"},
                    "operator": "=",
                },
            ],
            "boolean": "and",
        }
    },
    "order": [
        {"column_name": "customer_count", "order": "asc"},
        {"column_name": "store_sales.customer.state", "order": "asc"},
    ],
    "limit": -1,
}


def test_validate_response_invalid_function_two():
    # check to make sure that the invalid function is detected
    filtering = FilterResultV2.model_validate(TEST_INVALID_FUNCTION_TWO["filtering"])
    columns = [
        Column.model_validate(x) for x in TEST_INVALID_FUNCTION_TWO["output_columns"]
    ]
    env = Environment()
    env.parse(
        """
key store_sales.store.id int;
key store_sales.customer.id int;
key store_sales.customer.state string;
key store_sales.store.state string;"""
    )
    response, ir = validate_response(
        environment=env,
        output_columns=columns,
        filtering=filtering,
        order=None,
        limit=100,
        prompt="shenanigans",
    )

    assert response["status"] == "invalid", response
    errors = response["errors"]
    assert "does not use a valid function" in str(errors), errors


INVALID_FIELD = """{
    "action": "submit_answer",
    "action_input": {
        "output_columns": [
            {"name": "store_sales.customer.state"},
            {"name": "customer_count", 
                "calculation": {
                    "operator": "COUNT",
                    "arguments": [
                        {"name": "store_sales.customer.id"}
                    ]
                }
            }
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": ">",
                        "left": {
                            "name": "store_sales.item.current_price"
                        },
                        "right": {
                            "value": {
                                "operator": "MULTIPLY",
                                "arguments": [
                                    {
                                        "name": "average_item_price_by_category"
                                    },
                                    {
                                        "value": "1.2",
                                        "type": "float"
                                    }
                                ]
                            },
                            "type": "float"
                        }
                    },
                    {
                        "operator": "=",
                        "left": {"name": "store_sales.date.date.year"},
                        "right": {"value": "2001", "type": "integer"}
                    },
                    {
                        "operator": "=",
                        "left": {"name": "store_sales.date.date.month"},
                        "right": {"value": "1", "type": "integer"}
                    },
                    {
                        "operator": "is not",
                        "left": {"name": "store_sales.item.category"},
                        "right": {"value": "null", "type": "null"}
                    },
                    {
                        "operator": ">=",
                        "left": {
                            "name": "customer_count"
                        },
                        "right": {
                            "value": "10",
                            "type": "integer"
                        }
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [
            {"column_name": "customer_count", "order": "asc nulls first"},
            {"column_name": "store_sales.customer.state", "order": "asc nulls first"}
        ],
        "limit": -1
    },
    "reasoning": "I corrected the syntax error in the filtering clause by ensuring that the right-hand side of the comparison is correctly formatted as a Calculation object with a specified type. Now, I will submit this response."
}"""


def test_validate_response_invalid_field():
    # check to make sure that the invalid function is detected
    PARSED_INVALID = json.loads(INVALID_FIELD)
    filtering = FilterResultV2.model_validate(
        PARSED_INVALID["action_input"]["filtering"]
    )
    columns = [
        Column.model_validate(x)
        for x in PARSED_INVALID["action_input"]["output_columns"]
    ]

    env = Environment()
    env.add_file_import(
        Path(__file__).parent / "tpc_ds_duckdb" / "store_sales.preql", "store_sales"
    )

    response, ir = validate_response(
        environment=env,
        output_columns=columns,
        filtering=filtering,
        order=None,
        limit=100,
        prompt="shenanigans",
    )

    assert response["status"] == "invalid", response
    errors = response["errors"]
    assert "is not a valid preexisting field returned by the get" in str(errors), errors


INVALID_CALCULATION = """{
        "output_columns": [
            {
                "name": "web_sales.item.name"
            },
            {
                "name": "web_sales.item.desc"
            },
            {
                "name": "web_sales.item.category"
            },
            {
                "name": "web_sales.item.class"
            },
            {
                "name": "web_sales.item.current_price"
            },
            {
                "name": "sum_external_sales_price",
                "calculation": {
                    "operator": "SUM",
                    "arguments": [
                        {
                            "name": "web_sales.external_sales_price"
                        }
                    ],
                    "over": [
                        {"name": "web_sales.item.name"},
                        {"name": "web_sales.item.desc"},
                        {"name": "web_sales.item.category"},
                        {"name": "web_sales.item.class"},
                        {"name": "web_sales.item.current_price"}
                    ]
                }
            },
            {
                "name": "class_revenue_ratio",
                "calculation": {
                    "operator": "MULTIPLY",
                    "arguments": [
                        {
                            "operator": "DIVIDE",
                            "arguments": [
                                {
                                    "name": "sum_external_sales_price"
                                },
                                {
                                    "name": "total_class_revenue",
                                    "calculation": {
                                        "operator": "SUM",
                                        "arguments": [
                                            {
                                                "name": "web_sales.external_sales_price"
                                            }
                                        ],
                                        "over": [
                                            {"name": "web_sales.item.class"}
                                        ]
                                    }
                                }
                            ]
                        },
                        {
                            "value": "100", 
                            "type": "float"
                        }
                    ]
                }
            }
        ],
        "filtering": {
            "root": {
                "values": [
                    {
                        "operator": "in",
                        "left": {"name": "web_sales.item.category"},
                        "right": {
                            "value": ["Sports", "Books", "Home"],
                            "type": "string"
                        }
                    },
                    {
                        "operator": ">=",
                        "left": {"name": "web_sales.date.date"},
                        "right": {"value": "1999-02-22", "type": "date"}
                    },
                    {
                        "operator": "<=",
                        "left": {"name": "web_sales.date.date"},
                        "right": {"value": "1999-03-24", "type": "date"}
                    }
                ],
                "boolean": "and"
            }
        },
        "order": [
            {"column_name": "web_sales.item.category", "order": "asc"},
            {"column_name": "web_sales.item.class", "order": "asc"},
            {"column_name": "web_sales.item.name", "order": "asc"},
            {"column_name": "web_sales.item.desc", "order": "asc"},
            {"column_name": "class_revenue_ratio", "order": "asc"}
        ],
        "limit": 100
    }"""


def test_validate_calculation():
    # check to make sure that the invalid function is detected
    PARSED_INVALID = json.loads(INVALID_CALCULATION)
    filtering = FilterResultV2.model_validate(PARSED_INVALID["filtering"])
    columns = [Column.model_validate(x) for x in PARSED_INVALID["output_columns"]]

    env = Environment()
    env.add_file_import(
        Path(__file__).parent / "tpc_ds_duckdb" / "web_sales.preql", "web_sales"
    )

    response, ir = validate_response(
        environment=env,
        output_columns=columns,
        filtering=filtering,
        order=None,
        limit=100,
        prompt="shenanigans",
    )

    assert response["status"] == "valid", response



# imports
import random
import re
import time

import openai
from trilogy.constants import logger
from trilogy.core.models import Concept

from trilogy_nlp.constants import DEFAULT_LIMIT

extract = re.compile("Please try again in ([0-9]+)s")


def safe_limit(input: int | None) -> int:
    if not input:
        return DEFAULT_LIMIT
    if input in (-1, 0):
        return DEFAULT_LIMIT
    return input


def is_relevent_concept(concept: Concept) -> bool:
    if concept.name.startswith("_"):
        return False
    if concept.name.endswith(".count"):
        return False
    if concept.namespace and any(
        x.startswith("_") for x in concept.namespace.split(".")
    ):
        return False
    return True


# define a retry decorator
def retry_with_exponential_backoff(
    func,
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 10,
    errors: tuple = (openai.RateLimitError,),
):
    """Retry a function with exponential backoff."""

    def wrapper(*args, **kwargs):
        # Initialize variables
        num_retries = 0
        delay = initial_delay

        # Loop until a successful response or max_retries is hit or an exception is raised
        while True:
            try:
                return func(*args, **kwargs)

            # Retry on specified errors
            except errors as e:
                # Increment retries
                m = extract.search(str(e))

                # if we have a delay
                # wait for that
                # but continue to backoff if lots of retries
                # as there may be concurrent requests
                if extract.match(str(e)):
                    delay = int(m.group(1)) + 1
                    for x in range(num_retries):
                        delay *= exponential_base * (1 + jitter * random.random())
                else:
                    delay *= exponential_base * (1 + jitter * random.random())

                num_retries += 1

                # Check if max retries has been reached
                if num_retries > max_retries:
                    raise Exception(
                        f"Maximum number of retries ({max_retries}) exceeded."
                    )
                logger.info(f"Retrying on rate limit error for {delay}")
                # Increment the delay
                # delay *= exponential_base * (1 + jitter * random.random())

                # Sleep for the delay
                time.sleep(delay)

            # Raise exceptions for any errors not specified
            except Exception as e:
                raise e

    return wrapper



from typing import List

from trilogy.constants import DEFAULT_NAMESPACE
from trilogy.core.enums import (
    BooleanOperator,
    ComparisonOperator,
    FunctionClass,
    FunctionType,
    InfiniteFunctionArgs,
    Ordering,
    Purpose,
)
from trilogy.core.models import (
    AggregateWrapper,
    Comparison,
    Concept,
    ConceptTransform,
    Conditional,
    DataType,
    Environment,
    Function,
    HavingClause,
    MagicConstants,
    Metadata,
    OrderBy,
    OrderItem,
    Parenthetical,
    WhereClause,
)
from trilogy.core.processing.utility import (
    decompose_condition,
    is_scalar_condition,
)
from trilogy.parsing.common import arbitrary_to_concept, arg_to_datatype

from trilogy_nlp.llm_interface.constants import (
    MAGIC_GENAI_DESCRIPTION,
)
from trilogy_nlp.llm_interface.models import (
    Calculation,
    Column,
    FilterResultV2,
    Literal,
    NLPComparisonGroup,
    NLPConditions,
    OrderResultV2,
)


def parse_object(ob, environment: Environment):
    if isinstance(ob, Column):
        return create_column(ob, environment)
    elif isinstance(ob, Calculation):
        return create_anon_calculation(ob, environment)
    return create_literal(ob, environment)


def parse_datatype(dtype: str):
    mapping = {item.value.lower(): item for item in DataType}
    mapping["integer"] = DataType.INTEGER
    if dtype.lower() in mapping:
        return mapping[dtype]
    return DataType.STRING


def get_next_inline_calc_name(environment: Environment) -> str:
    return f"inline_calc_{len(environment.concepts)+1}"


def create_literal(
    literal: Literal, environment: Environment
) -> (
    str
    | float
    | int
    | bool
    | MagicConstants
    | Concept
    | ConceptTransform
    | list[str]
    | list[int]
    | list[float]
):
    # LLMs might get formats mixed up; if they gave us a column, hydrate it here.
    # and carry on
    if isinstance(literal.value, Calculation):
        return create_column(
            Column(
                name=get_next_inline_calc_name(environment), calculation=literal.value
            ),
            environment,
        )
    if isinstance(literal.value, str) and literal.value in environment.concepts:
        return create_column(Column(name=literal.value), environment)

    # otherwise, we really have a literal
    if literal.type == "null":
        return MagicConstants.NULL
    dtype = parse_datatype(literal.type)
    if isinstance(literal.value, list):
        return literal.value
    if dtype == DataType.STRING:
        return literal.value
    if dtype == DataType.INTEGER:
        return int(literal.value)
    if dtype == DataType.FLOAT:
        return float(literal.value)
    if dtype == DataType.BOOL:
        return bool(literal.value)
    if dtype == DataType.ARRAY:
        return literal.value
    return literal.value


def create_anon_calculation(
    c: Calculation, environment: Environment
) -> Function | AggregateWrapper:
    operator = FunctionType(c.operator.lower())
    if operator in FunctionClass.AGGREGATE_FUNCTIONS.value:
        purpose = Purpose.METRIC
    else:
        purpose = Purpose.PROPERTY
    if c.operator.lower() not in FunctionType.__members__:
        if c.operator == "RENAME":
            c.operator = "ALIAS"
    args = [parse_object(c, environment) for c in c.arguments]
    deriv_function = Function(
        operator=FunctionType(c.operator.lower()),
        output_datatype=arg_to_datatype(args[0]),
        output_purpose=purpose,
        arguments=args,
        arg_count=InfiniteFunctionArgs,
    )
    derivation: Function | AggregateWrapper
    if c.over:
        if purpose != Purpose.METRIC:
            raise ValueError("Can only use over with aggregate functions.")
        derivation = AggregateWrapper(
            function=deriv_function,
            by=[parse_object(c, environment) for c in c.over],
        )
    else:
        derivation = deriv_function
    return derivation


def create_column(c: Column, environment: Environment) -> Concept | ConceptTransform:
    if not c.calculation:
        return environment.concepts[c.name]
    base_name = c.name
    # TAG: resiliency
    # LLM may reference the same name for the output of a calculation
    # if that's so, force the outer concept a new name
    derivation = create_anon_calculation(c.calculation, environment)
    if any(
        isinstance(z, Concept) and z.name == base_name
        for z in derivation.concept_arguments
    ):
        base_name = f"{c.name}_deriv"
    new = arbitrary_to_concept(
        derivation,
        namespace=DEFAULT_NAMESPACE,
        name=f"{base_name}".lower(),
        metadata=Metadata(description=MAGIC_GENAI_DESCRIPTION),
        environment=environment,
    )
    environment.add_concept(new)

    return new


def parse_order(
    input_concepts: List[Concept | ConceptTransform],
    ordering: List[OrderResultV2] | None,
) -> OrderBy:
    normalized = [x if isinstance(x, Concept) else x.output for x in input_concepts]
    default_order = [
        OrderItem(expr=c, order=Ordering.DESCENDING)
        for c in normalized
        if c.purpose == Purpose.METRIC
    ]
    if not ordering:
        return OrderBy(items=default_order)
    final = []
    for order in ordering:
        possible_matches = [
            x
            for x in normalized
            if x.address == order.column_name or x.name == order.column_name
        ]
        if not possible_matches:

            possible_matches = [
                x
                for x in normalized
                if x.lineage
                and any(
                    [
                        y.address == order.column_name
                        for y in x.lineage.concept_arguments
                    ]
                )
            ]
        if possible_matches:
            final.append(OrderItem(expr=possible_matches[0], order=order.order))
    return OrderBy(items=final)


def parse_filter_obj(
    inp: NLPComparisonGroup | NLPConditions | Column | Literal | list[Column | Literal],
    environment: Environment,
):
    if isinstance(inp, NLPComparisonGroup):
        children = [parse_filter_obj(x, environment) for x in inp.values]

        def generate_conditional(list: list, operator: BooleanOperator):
            left = list.pop(0)
            if not list:
                return left
            right = generate_conditional(list, operator)
            return Conditional(left=left, right=right, operator=operator)

        base = generate_conditional(children, operator=inp.boolean)
        if inp.boolean == BooleanOperator.OR:
            return Parenthetical(content=base)
        return base
    elif isinstance(inp, NLPConditions):
        left = parse_filter_obj(inp.left, environment)
        right = parse_filter_obj(inp.right, environment)
        operator = inp.operator
        if right == MagicConstants.NULL and operator == ComparisonOperator.NE:
            operator = ComparisonOperator.IS_NOT
        cast_eligible = operator not in [
            ComparisonOperator.IS_NOT,
            ComparisonOperator.IS,
            ComparisonOperator.IN,
            ComparisonOperator.NOT_IN,
        ]
        if cast_eligible and arg_to_datatype(left) != arg_to_datatype(right):
            right = Function(
                operator=FunctionType.CAST,
                output_datatype=arg_to_datatype(left),
                output_purpose=Purpose.PROPERTY,
                arguments=[right, arg_to_datatype(left)],  # type: ignore
                arg_count=2,
            )
        return Comparison(
            left=left,
            right=right,
            operator=operator,
        )
    elif isinstance(inp, (Column, Literal)):
        return parse_object(inp, environment)
    elif isinstance(inp, list):
        return [parse_filter_obj(x, environment) for x in inp]
    else:
        raise SyntaxError(inp)


def parse_filter_obj_flat(
    inp: NLPComparisonGroup | NLPConditions | Column | Literal, environment: Environment
):
    if isinstance(inp, NLPComparisonGroup):
        children = [parse_filter_obj(x, environment) for x in inp.values]

        def generate_conditional(list: list, operator: BooleanOperator):
            if not list:
                return True
            left = list.pop(0)
            right = generate_conditional(list, operator)
            flat = [
                left,
            ] + right
            if operator == BooleanOperator.AND:
                return flat
            else:
                return [flat]

        return generate_conditional(children, operator=inp.boolean)
    elif isinstance(inp, NLPConditions):
        return Comparison(
            left=parse_filter_obj(inp.left, environment),
            right=parse_filter_obj(inp.right, environment),
            operator=inp.operator,
        )
    elif isinstance(inp, (Column, Literal)):
        return parse_object(inp, environment)
    else:
        raise SyntaxError(inp)


def parse_filter(
    input: FilterResultV2, environment: Environment
) -> Comparison | Conditional | None:
    if not input.root.values:
        return None
    return parse_filter_obj(input.root, environment)


def parse_filtering(
    filtering: FilterResultV2, environment: Environment
) -> WhereClause | None:
    parsed = parse_filter(filtering, environment=environment)
    if parsed:
        return WhereClause(conditional=parsed)
    return None


def generate_having_and_where(
    selected: list[str],
    filtering: WhereClause | None = None,
) -> tuple[WhereClause | None, HavingClause | None]:
    """If a concept is output by the select, and it is an aggregate;
    move that condition to the having clause. If it's an aggregate that is not output,
    we can keep it in where, as well as any scalars."""
    if not filtering:
        return None, None
    where: Conditional | Comparison | Parenthetical | None = None
    having: Conditional | Comparison | Parenthetical | None = None

    if is_scalar_condition(filtering.conditional):
        where = filtering.conditional
    else:
        components = decompose_condition(filtering.conditional)
        for x in components:
            if is_scalar_condition(x):
                where = where + x if where else x
            else:
                if any([c.address in selected for c in x.concept_arguments]):
                    if any(
                        (
                            (
                                isinstance(x.lineage, AggregateWrapper)
                                # and not x.lineage.by
                            )
                            or (
                                isinstance(x.lineage, Function)
                                and x.lineage.operator
                                in FunctionClass.AGGREGATE_FUNCTIONS.value
                            )
                        )
                        for x in x.concept_arguments
                    ):
                        having = having + x if having else x
                        continue
                # otherwise we end up here
                where = where + x if where else x
    return WhereClause(conditional=where) if where else None, (
        HavingClause(conditional=having) if having else None
    )



import os
from datetime import datetime
from pathlib import Path as PathlibPath
from sys import path as sys_path

from click import UNPROCESSED, Path, argument, group
from trilogy.dialect.enums import Dialects  # noqa
from trilogy.executor import Executor
from trilogy.parsing.render import Renderer

from trilogy_nlp.environment import build_env_and_imports
from trilogy_nlp.main import build_query

# handles development cases
nb_path = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
sys_path.insert(0, nb_path)

renderer = Renderer()


def generate_executor(dialect):
    pass


def print_tabulate(q, tabulate):
    result = q.fetchall()
    print(tabulate(result, headers=q.keys(), tablefmt="psql"))


def run_query(
    text: str, working_path: PathlibPath, dialect: Dialects, engine: Executor, llm
):
    engine = generate_executor(dialect)
    env = build_env_and_imports(text, working_path=working_path, llm=llm)
    processed_query = build_query(
        input_text=text,
        input_environment=env,
        debug=True,
        llm=llm,
    )
    # fetch our results
    parse_start = datetime.now()
    engine.environment = env
    query = engine.generate_sql(processed_query)[-1]
    parse_time = datetime.now() - parse_start
    print(f"Parse time: {parse_time}")
    exec_start = datetime.now()
    results = engine.execute_raw_sql(query)
    print(f"Exec time: {datetime.now() - exec_start}")
    if not results:
        print("Empty results")
    try:
        import tabulate

        print_tabulate(results, tabulate.tabulate)
    except ImportError:
        print("Install tabulate (pip install tabulate) for a prettier output")
        print(", ".join(results.keys()))
        for row in results:
            print(row)
        print("---")
    return query


@group()
def main():
    """Parent CLI"""
    pass


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@argument("preql", type=Path())
@argument("output_path", type=Path(exists=False))
@argument("dialect", type=str)
@argument("conn_args", nargs=-1, type=UNPROCESSED)
def simple(preql: str | Path, output_path: Path, dialect: str, conn_args):
    # _= Dialects(dialect)
    _: PathlibPath = PathlibPath(str(preql))


if __name__ == "__main__":
    main()



from os.path import dirname
from sys import path

nb_path = __file__
root_path = dirname(dirname(dirname(__file__)))

print(root_path)
path.insert(0, root_path)


from logging import DEBUG, StreamHandler

from trilogy.core.models import Environment

from examples.titanic.setup_environment import setup_engine, setup_titanic
from trilogy_nlp.constants import logger
from trilogy_nlp.core import NLPEngine
from trilogy_nlp.enums import Provider
from trilogy_nlp.main import build_query

# how many passengers survived in first and second class?

TEST_CASES = [
    "how many passengers survived in first and second class?",
    "HOw many passengers survived in each cabin?",
]
if __name__ == "__main__":
    provider_type = Provider.GOOGLE
    executor = setup_engine()

    env = Environment()
    model = setup_titanic(env)
    for c in env.concepts:
        print(c)

    executor.environment = env
    environment = env
    logger.setLevel(DEBUG)
    logger.addHandler(StreamHandler())

    open_ai = NLPEngine(provider=Provider.OPENAI, model="gpt-3.5-turbo")
    open_ai.test_connection()

    question = ("HOw many passengers survived in each cabin?",)
    processed_query = build_query(
        question,
        environment,
        debug=True,
        llm=open_ai.llm,
    )

    google_engine = NLPEngine(provider=Provider.GOOGLE)

    google_engine.test_connection()
    processed_query_v2 = build_query(
        question, environment, debug=True, llm=NLPEngine(provider=Provider.GOOGLE).llm
    )

    for q in [processed_query, processed_query_v2]:
        print("output_columns")
        for key in q.output_columns:
            print(key)

        print(executor.generator.compile_statement(q))
        results = executor.execute_query(q)
        for row in results:
            print(row)



FILTERING_EXAMPLE = {
    "root": {
        "boolean": "and",
        "values": [
            {
                "operator": ">",
                "left": {
                    "name": "avg_monthly_rainfall",
                    "calculation": {
                        "operator": "AVG",
                        "arguments": [{"name": "country.monthly_rainfall"}],
                        "over": [{"name": "country.name"}],
                    },
                },
                "right": {
                    "name": "continent_avg_monthly_rainfall_2x",
                    "calculation": {
                        "operator": "MULTIPLY",
                        "arguments": [
                            {"value": "2", "type": "integer"},
                            {
                                "name": "continent_avg_monthly_rainfall",
                                "calculation": {
                                    "operator": "AVG",
                                    "arguments": ["country.monthly_rainfall"],
                                    "over": [{"name": "country.continent"}],
                                },
                            },
                        ],
                    },
                },
            }
        ],
    }
}

COLUMN_DESCRIPTION = """A Column Object is json with two fields:
    -- name: the field being referenced or a new derived name. If there is a calculation, this should always be a new derived name you came up with. 
    -- calculation: An optional calculation object.
"""



from langchain.globals import set_llm_cache
from trilogy import Environment, Executor
from trilogy.core.models import ProcessedQuery
from trilogy.executor import CursorResult

from trilogy_nlp.enums import CacheType, Provider
from trilogy_nlp.instrumentation import EventTracker
from trilogy_nlp.main import build_query

DEFAULT_GPT = "gpt-4o-mini"
DEFAULT_GEMINI = "gemini-pro"

DEFAULT_MAX_TOXENS = 6500

# 0 to 1.0, scaled by provider
DEFAULT_TEMPERATURE = 0.25


class NLPEngine(object):
    def __init__(
        self,
        provider: Provider,
        model: str | None = None,
        api_key: str | None = None,
        cache: CacheType | None = None,
        cache_kwargs: dict | None = None,
        instrumentation: EventTracker | None = None,
    ):
        self.provider = provider
        self.debug = False
        self.model = model
        self.api_key = api_key
        self.llm = self.create_llm()
        self.cache = self.create_cache(cache, cache_kwargs or {}) if cache else None
        self.instrumentation = instrumentation

    def create_cache(self, cache: CacheType, cache_kwargs: dict):
        from langchain_core.caches import BaseCache

        cache_instance: BaseCache
        if cache == CacheType.SQLLITE:
            from langchain_community.cache import SQLiteCache

            cache_instance = SQLiteCache(**cache_kwargs)
        elif cache == CacheType.MEMORY:
            from langchain_community.cache import InMemoryCache

            cache_instance = InMemoryCache()
        else:
            return None
        set_llm_cache(cache_instance)
        return cache_instance

    def create_llm(self):
        if self.provider == Provider.OPENAI:
            import openai
            from langchain_openai import ChatOpenAI

            llm = ChatOpenAI(
                model_name=self.model if self.model else DEFAULT_GPT,
                openai_api_key=self.api_key,
                # openai temperature is 0 to 2.0
                temperature=DEFAULT_TEMPERATURE * 2,
                # model_kwargs=chat_openai_model_kwargs,
            ).with_retry(
                retry_if_exception_type=(
                    openai.APIConnectionError,
                    openai.APIError,
                    openai.APITimeoutError,
                )
            )
        elif self.provider == Provider.GOOGLE:
            from langchain_google_genai import ChatGoogleGenerativeAI

            llm = ChatGoogleGenerativeAI(
                model=self.model if self.model else DEFAULT_GEMINI,
                convert_system_message_to_human=True,
                google_api_key=self.api_key,
                # openai temperature is 0 to 2.0
                temperature=DEFAULT_TEMPERATURE * 2,
            ).with_retry()
        elif self.provider == Provider.LLAMAFILE:
            from langchain_community.chat_models import ChatOpenAI

            llm = ChatOpenAI(
                model_name=self.model if self.model else "not-applicalbe",
                openai_api_key="not-required",
                base_url="http://localhost:8080/v1",
                temperature=DEFAULT_TEMPERATURE,
                # model_kwargs=chat_openai_model_kwargs,
            ).with_retry()
        else:
            raise NotImplementedError(f"Unsupported provider {self.provider}")
        return llm

    def test_connection(self):
        local = self.create_llm()
        result = local.invoke("Hello")
        print(result.content)

    def generate_query(self, text: str, env: Environment | Executor) -> ProcessedQuery:
        if isinstance(env, Executor):
            env = env.environment
        # avoid mutating our model
        env = env.model_copy(deep=True)
        return build_query(
            input_text=text,
            input_environment=env,
            debug=self.debug,
            llm=self.llm,
        )

    def run_query(self, text: str, executor: Executor) -> CursorResult:
        query = self.generate_query(text, executor)
        return executor.execute_query(query)
